#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
annotate is a subcommand of PanACoTA

It is a pipeline to do quality control and annotate genomes. Steps are:

- optional: find rows of at least 'n' N (default n=5), and cut into a new contig at this point
- for each genome, calc L90 and number of contigs (after cut at n 'N' occurrences if used)
- keep only genomes with:

    - L90 <= x (default x = 100)
    - #contig <= y (default y = 999)
- rename those genomes and their contigs, with strain name increasing with quality (L90 and
  #contig)
- annotate kept genomes with prokka (default) or only prodigal
- gembase format

Input:

- **list_file:** list of genome filenames to annotate. This file contains 1 line per genome. It
  contains the name(s) of the (multi-)fasta file(s) corresponding to the genome (separated
  by space if several fasta files for the genome). After quality control, selected genomes
  will be named as following: ``<gen-spe>.<date>.<strain>``, with:

    * ``<gen_spe>`` 4 alphanumeric characters. Usually it corresponds to the 2 first letters
      of genus, and 2 first letters of species, like ESCO for Escherichia coli.
    * ``<date>`` date at which the genome was downloaded, formatted as MMYY (M=Month, Y=Year)
    * ``<strain>`` is the strain number of the genome in the species, ordered by quality.

Default values for ``<gen_spe>`` and ``<date>`` are given as input (see after). However, if some
genomes do not have the same date and/or genus/species as the others, you can add
this information for those genomes in the list file. fasta filenames and information are
separated by ``::``. ``<gen_spe>`` is given after the ``::``, and ``<date>`` is preceded by a
``.``. Here is an example::

    genome1.fasta
    genome2_ch1.fna genome2_pl.fst
    genome3.fst genome3_plasmid.fst :: name
    genome4.fna genome4.p1.fna genome4.p2.fna :: name.
    genome5.fasta :: name.date
    genome6.chromo.fst genome6.pl.fst  :: .date

- **species:** with 4 alphanumeric characters, used to rename genomes (except those whose
  species name is specified in the list file)
- **date:** optional. By default, takes the current date. Used to rename genomes (except those
  whose date is specified in the list file)
- **dbpath:** path to folder containing all multi-fasta sequences of genomes
- **respath:** path to folder where outputs must be saved (folders Genes, Replicons, Proteins,
  LSTINFO, gff3 and LSTINFO_dataset.lst file)
- **tmppath** optional. Path where tmp files must be saved. Default is respath/tmp_files
- **annotepath** optional. Path where prokka/prodigal output folders for all genomes must be saved.
  Default is respath/tmp_files
- **threads:** number of threads that can be used (default 1)

Output:

- In your given ``respath``, you will find 5 folders:

    * LSTINFO (information on each genome, with gene annotations),
    * Genes (nuc. gene sequences),
    * Proteins (aa proteins sequences),
    * Replicons (input sequences but with formatted headers).
    * gff3 (information on genes as gff3 format)

- In your given ``tmppath`` folder, folders with prokka/prodigal results will
  be created for each input genome (1 folder per genome, called
  ``<genome_name>-[prokka, prodigal]Res``). If errors are generated during prokka/prodigal
  step, you can look at the log file to see what was wrong
  (``<genome_name>-[prokka, prodigal].log``).
- In your given ``respath``, a file called ``annotate-genomes-<list_file>.log`` will be generated.
  You can find there all logs.
- In your given ``respath``, a file called ``annotate-genomes-<list_file>.log.err`` will be
  generated, containing information on errors and warnings that occurred: problems during
  annotation (hence no formatting step ran), and problems during formatting step. If this file is
  empty, then annotation and formatting steps finished without any problem for all genomes.
- In your given ``respath``, you will find a file called ``LSTINFO-<list_file>.lst`` with
 information on all genomes: gembase_name, original_name, genome_size, L90, nb_contigs
- In your given ``respath``, you will find a file called ``discarded-<list_file>.lst`` with
  information on genomes that were discarded (and hence not annotated) because of the
  L90 and/or nb_contig threshold: original_name, genome_size, L90, nb_contigs
- In your given ``respath``, you will find 2 png files: ``QC_L90-<list_file>.png`` and
  ``QC_nb-contigs-<list_file>.png``, containing the histograms of L90 and nb_contigs values for
  all genomes, with a vertical red line representing the limit applied here.

Requested:

- in prokka/prodigal results, all genes are called ``<whatever>_<number>``
   -> the number will be kept.
- The number of the genes annotated by prokka/prodigal are in increasing order
  in tbl, faa and ffn files
- genome names given to prokka/prodigal should not end with '_<number>'. Ideally, they should
  always have the same format: ``<spegenus>.<date>.<strain_number>`` but they can have
  another format, as long as they don't end by '_<number>', which is the format of a gene name.

@author gem
April 2017
"""

import os
import sys
from termcolor import colored
import sys


def main_from_parse(arguments):
    """
    Call main function from the arguments given by parser

    Parameters
    ----------
    arguments : argparse.Namespace
        result of argparse parsing of all arguments in command line

    """
    cmd = "PanACoTA " + ' '.join(arguments.argv)
    main(cmd, arguments.list_file, arguments.db_path, arguments.res_path,
         arguments.name,
         arguments.date, arguments.l90, arguments.nbcont, arguments.cutn, arguments.threads,
         arguments.force, arguments.qc_only, arguments.from_info, arguments.tmpdir,
         arguments.annotdir, arguments.verbose, arguments.quiet, arguments.prodigal_only,
         arguments.small)


def main(cmd, list_file, db_path, res_dir, name, date, l90=100, nbcont=999, cutn=5,
         threads=1, force=False, qc_only=False, from_info=None, tmp_dir=None, res_annot_dir=None,
         verbose=0, quiet=False, prodigal_only=False, small=False):
    """
    Main method, doing all steps:

    1. analyze genomes (nb contigs, L90, rows of N...)
    2. keep only genomes with 'good' (according to user thresholds) L90 and nb_contigs
    3. rename genomes with strain number in decreasing quality
    4. annotate genome with prokka or only prodigal
    5. format annotated genomes

    If option '-Q': ends at step 2.
    If option '--info <genome_info file name>' option: starts at step 2

    verbosity:

    - defaut 0 : stdout contains INFO, stderr contains ERROR.
    - 1: stdout contains INFO, stderr contains WARNING and ERROR
    - 2: stdout contains (DEBUG), DETAIL and INFO, stderr contains WARNING and ERROR
    - >=15: Add DEBUG in stdout

    Parameters
    ----------
    cmd : str
        command line used to launch this program
    list_file : str
        file containing the list of genome files, 1 genome per line, separated by a
        space if a genome is split in several fasta files. This file can also
        specify date and/or species information, according to the format described
        in documentation.
    db_path : str
        Path to the folder containing all the fasta files which will be annotated
    res_dir : str
        Path to the folder which will contain result folders and files
    name : str
        4 alpha numeric characters, describing the species (for example ESCO). Used by default
        if no species name is given in list_file line.
    date : str
        4 alpha numeric characters, defining the default date, for strains where it is not specified
        in the list_file
    l90 : int
        Max L90 allowed to keep a genome
    nbcont : int
        Max number of contigs allowed to keep a genome
    cutn : int
        cut each time there are at least cutn 'N' in a row. Don't cut if equal to 0
    threads : int
        max number of threads to use
    force : bool
        If True, overwrite previous results, if False keep what is already calculated
    qc_only : bool
        If True, do only quality control, if False, also do annotation
    from_info : str
        File containing information on genomes and their quality information (from prepare step)
    tmp_dir : str or None
        Path to folder where tmp files must be saved. None to use the default tmp folder
    res_annot_dir : str or None
        Path to folder where are the prokka/prodigal result folders for the genomes. None
        to use the default prokka/prodigal folder
    verbose : int
        verbosity:
        default (0): info in stdout, error and more in stderr
        1 = add warnings in stderr
        2 = like 1 + add DETAIL to stdout (by default only INFO)
        >15: add debug to stdout
    quiet : bool
        True if nothing must be sent to stdout/stderr, False otherwise
    prodigal_only : bool
        True -> run only prodigal. False -> run prokka
    small : bool
        True -> use -p meta option with prodigal

    Returns
    -------
    (genomes, kept_genomes, skipped, skipped_format) : tuple
        with:

        - genomes: dict with all genomes in list_file:
          {genome: [gembase_name, path_split_gembase, gsize, nbcont, L90]}
        - kept_genomes: dict with all genomes kept for annotation (same format as genomes)
        - skipped: list of genomes skipped because they had a problem in annotation step
        - skipped_format : list of genomes skipped because they had a problem in format step
    """
    # import needed packages
    import shutil
    import logging
    from PanACoTA.annotate_module import genome_seq_functions as gfunc
    from PanACoTA.annotate_module import annotation_functions as pfunc
    from PanACoTA.annotate_module import general_format_functions as ffunc
    from PanACoTA import utils
    from PanACoTA import __version__ as version
    # Check that needed softs are installed
    prokka = utils.check_installed("prokka")
    prodigal = utils.check_installed("prodigal")
    if prodigal_only:
        soft = "prodigal"
    else:
        soft = "prokka"

    changed = cutn != 0
    if not qc_only: # pragma: no cover
        # If user using prokka: check prokka is installed and in the path
            if not prodigal_only and not prokka:
                print("Prokka is not installed. 'PanACoTA annotate' cannot run. Install prokka "
                      "to be able to annotate genomes. If you only need syntactical annotation, "
                      "check that prodigal is installed, and add '--prodigal' option.")
                sys.exit(1)
            if prodigal_only and not prodigal:
                print("Prodigal is not installed. 'PanACoTA annotate' cannot run. Install "
                      "prodigal to be able to annotate genomes. If you also need functional "
                      "annotation, check that prokka is installed, and remove '--prodigal' "
                      "option.")
                sys.exit(1)

    # By default, all tmp files (split sequences, renamed sequences, prokka/prodigal results) will
    # be saved in the given <res_dir>/tmp_files.
    # Create output (results, tmp...) directories if not already existing
    if not tmp_dir:
        tmp_dir = os.path.join(res_dir, "tmp_files")
    if not res_annot_dir:
        res_annot_dir = tmp_dir
    os.makedirs(res_dir, exist_ok=True)
    os.makedirs(tmp_dir, exist_ok=True)
    os.makedirs(res_annot_dir, exist_ok=True)

    # If force was set, remove result folders (Proteins, Replicons, Genes, LSTINFO, gff)
    if force:
        shutil.rmtree(os.path.join(res_dir, "LSTINFO"), ignore_errors=True)
        shutil.rmtree(os.path.join(res_dir, "Proteins"), ignore_errors=True)
        shutil.rmtree(os.path.join(res_dir, "Genes"), ignore_errors=True)
        shutil.rmtree(os.path.join(res_dir, "Replicons"), ignore_errors=True)
        shutil.rmtree(os.path.join(res_dir, "gff3"), ignore_errors=True)
    # If not --force, check that result folders do not already contain results
    else:
        utils.check_out_dirs(res_dir)

    # get only filename of list_file, without extension
    if list_file:
        listfile_base = os.path.basename(os.path.splitext(list_file)[0])
    else:
        list_file = from_info
        listfile_base = os.path.basename(os.path.splitext(list_file)[0])

    # Initialize logger
    # set level of logger: level is the minimum level that will be considered.
    if verbose <= 1:
        level = logging.INFO
    # for verbose = 2, ignore only debug
    if verbose >= 2 and verbose < 15:
        level = utils.detail_lvl() # int corresponding to detail level
    # for verbose >= 15, write everything
    if verbose >= 15:
        level = logging.DEBUG
    logfile_base = os.path.join(res_dir, "PanACoTA-annotate_" + listfile_base)
    logfile_base = utils.init_logger(logfile_base, level, name='annotate', log_details=True,
                                     verbose=verbose, quiet=quiet)
    logger = logging.getLogger('annotate')
    logger.info(f'PanACoTA version {version}')
    logger.info("Command used\n \t > " + cmd)

    # STEP 1. analyze genomes (nb contigs, L90, rows of N...)
    # If already info on genome ('--info <file>' option), skip this step
    # If no info on genomes, read them and get needed information
    if not from_info:
        # Read genome names.
        # genomes = {genome: [spegenus.date]}
        genomes = utils.read_genomes(list_file, name, date, db_path, tmp_dir, logger)
        if not genomes:
            logger.error(("We did not find any genome listed in {} in the folder {}. "
                          "Please check your list to give valid genome "
                          "names.").format(list_file, db_path))
            sys.exit(1)
        # Get L90, nbcontig, size for all genomes, and cut at row of cutn 'N' if asked
        # -> genome: [spegenus.date, orig_path, to_annotate_path, size, nbcont, l90]
        gfunc.analyse_all_genomes(genomes, db_path, tmp_dir, cutn, soft,
                                  logger, quiet=quiet)
    # --info <filename> option given: read information (L90, nb contigs...) from this file.
    else:
        # genomes = {genome: [spegenus.date, orig_path, to_annotate_path, size, nbcont, l90]}
        # orig_path is the path to the original sequence
        # and to_annotate_path the path to the sequence to annotate (once split etc.)
        # Here, both are the same, as we take given sequences as is.
        genomes = utils.read_genomes_info(from_info, name, date, logger)

    # STEP 2. keep only genomes with 'good' (according to user thresholds) L90 and nb_contigs
    # genomes = {genome: [spegenus.date, orig_seq, path_to_splitSequence, size, nbcont, l90]}
    # Plot L90 and nb_contigs distributions
    gfunc.plot_distributions(genomes, res_dir, listfile_base, l90, nbcont)
    # Get list of genomes kept (according to L90 and nbcont thresholds)
    kept_genomes = {genome: info for genome, info in genomes.items()
                    if info[-2] <= nbcont and info[-1] <= l90}
    # Write discarded genomes to a file -> orig_name, to_annotate, gsize, nb_conts, L90
    utils.write_genomes_info(genomes, list(kept_genomes.keys()), list_file, res_dir)

    if not kept_genomes:
        logger.info("No genome kept for annotation.")
        return "", 0
    # Info on folder containing original sequences
    if not from_info:
        logger.info(f"-> Original sequences folder ('orig_name' column): {db_path} ")
        logger.info(f"\t-> If original sequence not found in {db_path}, "
                    f"look for it in {tmp_dir}, as it must be a concatenation of several "
                    "input sequence files.")
        if cutn == 0:
            logger.info("-> Sequences used for annotation ('to_annotate' column) are the "
                        "same as the previous ones (original sequences).")
        else:
            logger.info(f"-> Folder with sequence files that will be used for annotation "
                        f"('to_annotate' column): {tmp_dir}")
    # If only QC, stop here.
    if qc_only:
        # Write information on genomes that would be annotated with the current
        # parameters if not QC_only:
        # orig_name, to_annnote, gsize, nb_conts, L90
        utils.write_genomes_info(genomes, [], list_file, res_dir, qc=True)
        logger.info("QC only done.")
        return "", 0

    # STEP 3. Rename genomes kept, ordered by decreasing quality
    first_gname = gfunc.rename_all_genomes(kept_genomes)
    # kept_genomes = {genome: [gembase_name, path_to_origfile, path_split_gembase,
    #                 gsize, nbcont, L90]}
    # first_gname = name of the first genome
    # Write lstinfo file (list of genomes kept with info on L90 etc.)
    outlst = utils.write_lstinfo(list_file, kept_genomes, res_dir)

    # STEP 4. Annotate all kept genomes
    results = pfunc.run_annotation_all(kept_genomes, threads, force, res_annot_dir, first_gname,
                                       prodigal_only, small=small, quiet=quiet)
    # Information on genomes to format
    # results_ok = {genome: [gembase_name, path_to_origfile, path_split_gembase,
    #               gsize, nbcont, L90]}
    results_ok = {genome:info for genome, info in kept_genomes.items() if results[genome]}
    # If no genome was ok, no need to format them. Just print that no genome was annotated,
    # end program.
    if not results_ok:
        logger.error("Error: No genome was correctly annotated, no need to format them.")
        sys.exit(1)
    # list of genomes skipped because annotation had problems: no format step run
    skipped = [genome for (genome, ok) in results.items() if not ok]
    # At least 1 genome was not annotated: write a message to warn on it
    if skipped:
        utils.write_warning_skipped(skipped, prodigal_only=prodigal_only,
                                    logfile=logfile_base)

    # STEP 5. Format genomes annotated
    # Here, we have at least 1 genome annotated (otherwise,
    # it would already have stopped because results_ok is empty)
    # Initialize list of genomes skipped because something went wrong while formatting.
    skipped_format = []
    # Generate database (folders Proteins, Genes, Replicons, LSTINFO)
    skipped_format = ffunc.format_genomes(results_ok, res_dir, res_annot_dir,
                                          prodigal_only, threads, quiet=quiet)
    # At least one genome could not be formatted -> warn user
    if skipped_format:
        utils.write_warning_skipped(skipped_format, do_format=True, prodigal_only=prodigal_only,
                                    logfile = logfile_base)
    logger.info("Annotation step done.")
    return outlst, len(kept_genomes) - len(skipped) - len(skipped_format)


def build_parser(parser):
    """
    Method to create a parser for command-line options

    Parameters
    ----------
    parser : argparse.ArgumentParser
        The parser to configure

    """
    from PanACoTA import utils
    from PanACoTA import utils_argparse
    import multiprocessing
    import argparse

    # Create command-line parser for all options and arguments to give
    required = parser.add_argument_group('Required arguments')
    required.add_argument("-d", dest="db_path",
                          help="Path to folder containing all multifasta genome files")
    required.add_argument("-r", dest="res_path", required=True,
                          help="Path to folder where output annotated genomes must be saved")
    optional = parser.add_argument_group('Optional arguments')
    optional.add_argument('-l', dest="list_file",
                          help=("File containing the list of genome filenames to annotate "
                                "(1 genome per line). Each genome must be in multi-fasta format. "
                                "You can specify the species name (4 characters) you want to "
                                "give to some genome, as well as the download (or any other "
                                "reason) date of your choice. Format 'genome_name :: name.date'. "
                                "name and date are optional. See doc for more information on "
                                "this file format. "
                                "If you want to run this module from 'prepare_module' results, "
                                "use '--info' instead."))
    optional.add_argument("-n", dest="name", type=utils_argparse.gen_name,
                          help=("Choose a name for your annotated genomes. This name should "
                                "contain 4 alphanumeric characters. Generally, they correspond "
                                "to the 2 first letters of genus, and 2 first letters of "
                                "species, e.g. ESCO for Escherichia Coli."))
    optional.add_argument("-Q", dest="qc_only", action="store_true", default=False,
                          help="Add this option if you want only to do quality control on your "
                               "genomes (cut at 5N if asked, calculate L90 and number of contigs "
                               "and plot their distributions). This allows you to check which "
                               "genomes would be annotated with the given parameters, and to "
                               "modify those parameters if you want, before you launch the "
                               "annotation and formatting steps.")
    optional.add_argument("--info", dest="from_info",
                          help="If you already ran the 'prepare' data module, or already "
                               "calculated yourself the L90 and number of contigs for each "
                               "genome, you can give this information, to go directly to "
                               "annotation and formatting steps. This file contains at "
                               "least 4 columns, tab separated, with the following headers: "
                               "'to_annotate', 'gsize', 'nb_conts', 'L90'. Any other column "
                               "will be ignored.")
    optional.add_argument("--prodigal", dest="prodigal_only", action="store_true", default=False,
                          help="Add this option if you only want syntactical annotation, given "
                               "by prodigal, and not functional annotation requiring prokka and "
                               "is slower.")
    optional.add_argument("--small", dest="small", action="store_true", default=False,
                          help="If you use Prodigal to annotate genomes, if you sequences are "
                               "too small (less than 20000 characters), it cannot annotate them "
                               "with the default options. Add this option to use 'meta' procedure.")
    optional.add_argument("--l90", dest="l90", type=int, default=100,
                          help="Maximum value of L90 allowed to keep a genome. Default is 100.")
    optional.add_argument("--nbcont", dest="nbcont", type=utils_argparse.cont_num, default=999,
                          help=("Maximum number of contigs allowed to keep a genome. "
                                "Default is 999."))
    optional.add_argument("--cutn", dest="cutn", type=utils_argparse.positive_int, default=5,
                          help=("By default, each genome will be cut into new contigs when "
                                "at least 5 'N' in a row are found in its sequence. "
                                "If you don't want to "
                                "cut genomes into new contigs when there are rows of 'N', "
                                "put 0 to this option. If you want to cut from a different number "
                                "of 'N' occurrences, put this value to this option."))
    optional.add_argument("--date", dest="date", default=utils_argparse.get_date(),
                          type=utils_argparse.date_name,
                          help=("Specify the date (MMYY) to give to your annotated genomes. "
                                "By default, will give today's date. The only requirement on the"
                                " given date is that it is 4 characters long. You can use letters"
                                " if you want. But the common way is to use 4 digits, "
                                "corresponding to MMYY."))
    optional.add_argument("--tmp", dest="tmpdir",
                          help=("Specify where the temporary files (sequence split by rows "
                                "of 'N', sequence with new contig names etc.) must be saved. "
                                "By default, it will be saved in your "
                                "result_directory/tmp_files."))
    optional.add_argument("--annot_dir", dest="annotdir",
                          help=("Specify in which directory the prokka/prodigal output files "
                                "(1 folder per genome, called "
                                "<genome_name>-[prokka, Prodigal]Res) must be "
                                "saved. By default, they are saved in the same directory as "
                                "your temporary files (see --tmp option to change it)."))
    optional.add_argument("-F", "--force", dest="force", action="store_true",
                          help=("Force run: Add this option if you want to (re)run annotation and "
                                "formatting steps for all genomes "
                                "even if their result folder (for annotation step) or files (for "
                                "format step) already exist: override existing results.\n"
                                "Without this option, if there already are results in the given "
                                "result folder, the program stops. If there are no results, but "
                                "prokka/prodigal folder already exists, prokka/prodigal won't run "
                                "again, and the formating step will use the already existing "
                                "folder if correct, or skip the genome if there are problems in "
                                "prokka/prodigal folder."))
    optional.add_argument("--threads", dest="threads", type=utils_argparse.thread_num, default=1,
                          help="Specify how many threads can be used (default=1)")
    helper = parser.add_argument_group('Others')
    helper.add_argument("-v", "--verbose", dest="verbose", action="count", default=0,
                        help="Increase verbosity in stdout/stderr.")
    helper.add_argument("-q", "--quiet", dest="quiet", action="store_true", default=False,
                        help=("Do not display anything to stdout/stderr. log files will "
                              "still be created."))
    helper.add_argument("-h", "--help", dest="help", action="help",
                        help="show this help message and exit")


def parse(parser, argu):
    """
    arse arguments given to parser

    Parameters
    ----------
    parser : argparse.ArgumentParser
        the parser used
    argu : [str]
        command-line given by user, to parse using parser

    Returns
    -------
    argparse.Namespace
        Parsed arguments
    """
    import argparse

    args = parser.parse_args(argu)
    return check_args(parser, args)


def check_args(parser, args):
    """
    Check that arguments given to parser are as expected.

    Parameters
    ----------
    parser : argparse.ArgumentParser
        The parser used to parse command-line
    args : argparse.Namespace
        Parsed arguments

    Returns
    -------
    argparse.Namespace or None
        The arguments parsed, updated according to some rules. Exit program
        with error message if error occurs with arguments given.

    """
    # Message if user kept default thresholds for L90 and nbcont. Just to warn him, to be sure
    # it was on purpose
    def thresholds_message(l90, nbcont):
        return ("  !! Your genomes will be filtered, and only the ones with 'L90' <= {} "
                "and 'number of contigs' < {} will be kept. If you want to change those "
                "thresholds, use '--l90' and '--nbcont' "
                "options.".format(args.l90, args.nbcont))

    # Message if user is giving a file with already calculated information
    def nosplit_message():
        split = ("  !! Your sequences will be used as is by PanACoTA. Be sure you "
                 "already split your sequences at each row "
                 "of X 'N' if needed.\n")
        trust = ("\t-> PanACoTA will use the values (L90, nbcont) given in your info file. "
                 "It will ignore the genomes for which those values are incorrect. "
                 "It will also ignore genomes with more than 999 contigs.")
        return split + trust

    #  ERRORS
    # Cannot be verbose and quiet at the same time
    if args.verbose > 0 and args.quiet:
        parser.error("Choose between a verbose output (-v) or a quiet output (-q)."
                     " You cannot have both.")
    # User wants to run all annotation step: needs a genome dataset name
    if not args.qc_only and not args.name:
        parser.error("You must specify your genomes dataset name in 4 characters with "
                     "'-n name' option (type -h for more information). Or, if you do not want "
                     "to annotate and format your genomes but just to run quality control, use "
                     "option '-Q")
    # If QC only, we do not need name -> name is NONE
    if args.qc_only and not args.name:
        args.name = "NONE"
    # option --small used only with prodigal
    if not args.prodigal_only and args.small:
        parser.error("You cannot use --small option with prokka. Either use prodigal, "
                     "or remove this option.")
    # If user specifies a cutN value (different than default one which is 5), and give
    # an info file, it is not compatible: info file will use sequences as is, and won't cut them
    if args.cutn != 5 and args.from_info:
        parser.error("If you provide a list of genomes with their calculated L90 and number of "
                     "contigs, PanACoTA will use the given sequences as is. It will not cut "
                     "them. So, you cannot use both --cutn and --info.")

    # Give a lst_file or an info file, not nothing
    if not args.from_info and not args.list_file:
        parser.error("You must provide a list of genomes to annotate. Either raw genomes "
                     "(see -l option), or genomes with quality information (see --info option).")

    # Choose between infofile or LSTINFO
    if args.from_info and args.list_file:
        parser.error("Either you want to annotate raw sequences (name of files in '-l infofile') "
                     "which will first go through the QC process, "
                     "OR you already did QC on your sequences and just want to annotate them "
                     "(information on those sequences in '--info LSTINFO-file'). "
                     "Please choose one of these 2 possibilities.")

    # If no info file nor db_path, ask for 1 of them
    if not args.db_path and not args.from_info:
        parser.error("You must provide a path to your database genome sequences (-d <db_path>). "
                     "If you already have a LSTINFO file, it contains this db_path. Use it with "
                     "--info <lstinfo file> option.")

    # If given LSTINFO, already contains paths to genome to annotate. db_path must not be provided
    if args.from_info and args.db_path:
        parser.error("If you run from your LSTINFO file, this one already contains the "
                     "path of genomes to annotate. Remove -d <db_path> option.")

    # WARNINGS
    # If user wants to cut genomes, warn him to check that it is on purpose (because default is cut at each 5'N')
    if args.cutn != 0 and not args.from_info:
        message = ("  !! Your genomes will be split when sequence contains at "
                   "least {}'N' in a row. If you want to change this threshold, "
                   "see --cutn option.").format(args.cutn)
        print(colored(message, "yellow"))
    # Warn user about selection of genomes thresholds
    if args.l90 == 100 or args.nbcont == 999:
        print(colored(thresholds_message(args.l90, args.nbcont), "yellow"))
        if args.from_info:
            print(colored(nosplit_message(), "yellow"))
    print()
    return args

if __name__ == '__main__':
    import argparse
    from textwrap import dedent
    header = '''
     ___                 _____  ___         _____  _____
    (  _`\              (  _  )(  _`\      (_   _)(  _  )
    | |_) )  _ _   ___  | (_) || ( (_)   _   | |  | (_) |
    | ,__/'/'_` )/' _ `\|  _  || |  _  /'_`\ | |  |  _  |
    | |   ( (_| || ( ) || | | || (_( )( (_) )| |  | | | |
    (_)   `\__,_)(_) (_)(_) (_)(____/'`\___/'(_)  (_) (_)


       Large scale comparative genomics tools

     -------------------------------------------
     '''
    my_parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                        description=dedent(header), add_help=False)
    build_parser(my_parser)
    OPTIONS = parse(my_parser, sys.argv[1:])
    main_from_parse(OPTIONS)



#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
tree is a subcommand of PanACoTA

@author gem
June 2017
"""
import sys


def main_from_parse(args):
    """
    Call main function from the arguments given by parser

    Parameters
    ----------
    args : argparse.Namespace
        result of argparse parsing of all arguments in command line
    """
    cmd = "PanACoTA " + ' '.join(args.argv)
    main(cmd, args.alignment, args.outdir, args.soft, args.model, args.threads,
         args.boot, args.write_boot, args.memory, args.fast, args.verbose, args.quiet)


def main(cmd, align, outdir, soft, model, threads, boot=False, write_boot=False,
         memory=False, fast=False, verbose=0, quiet=False):
    """
    Inferring a phylogenetic tree from an alignment file, with the given software.

    Parameters
    ----------
    cmd: str
        command used to launch tree module
    align: str
        Path to file containing alignments of persistent families grouped by genome
    outdir: str or None
        Path to file which will contain the tree inferred
    soft: str
        Soft to use to infer the phylogenetic tree: 1 of quicktree, fasttree or fastme
    model: str or None
        DNA substitution model chosen by user, None if quicktree used
    threads: int
        Maximum number of threads to use
    boot: int or None
        Number of bootstraps to compute. None if no bootstrap asked
    write_boot: bool
        True if all bootstrap pseudo-trees must be saved into a file, False otherwise
    memory: str
        Maximal RAM usage in GB | MB | % - Only for iqtree
    fast: boolean
        use -fast option with IQtree
    verbose : int
        verbosity:
        - defaut 0 : stdout contains INFO, stderr contains ERROR.
        - 1: stdout contains INFO, stderr contains WARNING and ERROR
        - 2: stdout contains (DEBUG), DETAIL and INFO, stderr contains WARNING and ERROR
        - >=15: Add DEBUG in stdout
    quiet: bool
        True if nothing must be sent to stdout/stderr, False otherwise
    """
    # import needed packages
    import logging
    import os
    from PanACoTA import utils
    from PanACoTA import __version__ as version
    tree = None
    if soft == "fasttree":
        # test if fasttree is installed and in the path
        if not utils.check_installed("FastTreeMP"): # pragma: no cover
            print("FastTreeMP is not installed. 'PanACoTA tree' cannot run.")
            sys.exit(1)
        from PanACoTA.tree_module import fasttree_func as tree
    elif soft == "fastme":
        # test if fastME is installed and in the path
        if not utils.check_installed("fastme"): # pragma: no cover
            print("fastme is not installed. 'PanACoTA tree' cannot run.")
            sys.exit(1)
        from PanACoTA.tree_module import fastme_func as tree
    elif soft == "quicktree":
        # test if fastME is installed and in the path
        if not utils.check_installed("quicktree"):  # pragma: no cover
            print("quicktree is not installed. 'PanACoTA tree' cannot run.")
            sys.exit(1)
        from PanACoTA.tree_module import quicktree_func as tree
    elif soft == "iqtree2":
        # by default, iqtree2 (not iqtree).
        # So, if user did not specify, it means iqtree2. But if 'iqtree2' command
        # does not exist, use iqtree command instead.
        # test if iqtree2 is installed and in the path
        if not utils.check_installed("iqtree2"): # pragma: no cover
            if not utils.check_installed("iqtree"):
                print("IQtree2 is not installed. 'PanACoTA tree' cannot run.")
                sys.exit(1)
            else:
                soft == "iqtree"
        from PanACoTA.tree_module import iqtree_func as tree
    elif soft == "iqtree":
        # user specifically asked for iqtree (version 1)
        if not utils.check_installed("iqtree"): # pragma: no cover
            print("IQtree is not installed. 'PanACoTA tree' cannot run.")
            sys.exit(1)
        from PanACoTA.tree_module import iqtree_func as tree

    # If outdir does not already exist, create it
    if not os.path.isdir(outdir):
        os.makedirs(outdir)
    # name logfile, add timestamp if already existing
    logfile_base = os.path.join(outdir, "PanACoTA-tree-" + soft)
    # level is the minimum level that will be considered.
    # for verbose = 0 or 1, ignore details and debug, start from info
    if verbose <= 1:
        level = logging.INFO
    # for verbose = 2, ignore only debug
    if verbose >= 2 and verbose < 15:
        level = 15 # int corresponding to detail level
    # for verbose >= 15, write everything
    if verbose >= 15:
        level = logging.DEBUG

    utils.init_logger(logfile_base, level, 'tree', verbose=verbose, quiet=quiet, log_details=True)
    logger = logging.getLogger("tree")
    logger.info(f'PanACoTA version {version}')
    logger.info("Command used\n \t > " + cmd)
    tree.run_tree(align, boot, outdir, quiet, threads, model=model, wb=write_boot,
                  mem=memory, s=soft, f=fast)

    logger.info("END")


def build_parser(parser):
    """
    Method to create a parser for command-line options

    Parameters
    ----------
    parser : argparse.ArgumentParser
        parser to configure in order to extract command-line arguments
    """
    import argparse
    from PanACoTA import utils_argparse


    # Create command-line parser for all options and arguments to give
    required = parser.add_argument_group('Required arguments')
    required.add_argument("-a", dest="alignment", required=True,
                          help=("Alignment file in multi-fasta: each header will be a "
                                "leaf of the inferred tree."))
    required.add_argument("-o", dest="outdir", required=True,
                          help=("Directory where tree results will be saved. "))

    # Choose with which soft inferring phylogenetic tree
    softparse = parser.add_argument_group('Choose soft to use (default is IQtree2)')
    softs = ["fasttree", "fastme", "quicktree", "iqtree", "iqtree2"]
    softparse.add_argument("-s", "--soft", dest="soft", choices=softs, default="iqtree2",
                           help=("Choose with which software you want to infer the "
                                 "phylogenetic tree. Default is IQtree2 "
                                 "(versions 2.x of IQtree). If you want version 1.x of "
                                 "IQtree, use '-s iqtree'"))

    optional = parser.add_argument_group('Optional arguments')
    optional.add_argument("-b", "--boot", dest="boot", type=int,
                          help=("Indicate how many bootstraps you want to compute. By "
                                "default, no bootstrap is calculated. For IQtree, it "
                                "will use ultrafast bootstrap (>=1000)."))

    optional.add_argument("--threads", dest="threads", default=1, type=utils_argparse.thread_num,
                          help=("add this option if you want to parallelize on several threads. "
                                "Indicate on how many threads you want to parallelize. "
                                "By default, it uses 1 thread. Put 0 if you want to use "
                                "all threads of your computer. Not available with quicktree."))
    optional.add_argument("-m", "--model", dest="model",
                          help=("Choose your DNA substitution model.\n"
                                "Default for FastTree and IQtree: GTR. Default for FastME: TN93.\n"
                                "For FastTree, the choices are 'GTR' and 'JC'.\n"
                                "For FastME, choices are: 'p-distance' "
                                "(or 'p'), 'RY symmetric' (or 'Y'), 'RY' (or 'R'), "
                                "'JC69' (or 'J'), 'K2P' (or 'K'), 'F81' (or '1'), "
                                "'F84' (or '4'), 'TN93' (or 'T'), 'LogDet' (or 'L').\n"
                                "For IQtree, choices are HKY, JC, F81, K2P, K3P, K81uf,"
                                " TNef, TIM, TIMef, TVM, TVMef, SYM, GTR."))
    optional.add_argument("-B", dest="write_boot", action="store_true",
                          help=("Add this option if you want to write all bootstrap "
                                "pseudo-trees. Only available with FastME and IQtree."))
    optional.add_argument("--mem", dest="memory",
                          help=("Maximal RAM usage in GB | MB. Only available with iqtree."))
    optional.add_argument("-fast", dest="fast", action="store_true",
                          help=("Use -fast option with iqtree."))

    helper = parser.add_argument_group('Others')
    helper.add_argument("-v", "--verbose", dest="verbose", action="count", default=0,
                        help="Increase verbosity in stdout/stderr.")
    helper.add_argument("-q", "--quiet", dest="quiet", action="store_true", default=False,
                        help=("Do not display anything to stdout/stderr. log files will "
                              "still be created."))
    helper.add_argument("-h", "--help", dest="help", action="help",
                        help="show this help message and exit")


def check_args(parser, args):
    """
    Check that arguments given to parser are as expected.

    Parameters
    ----------
    parser : argparse.ArgumentParser
        The parser used to parse command-line
    args : argparse.Namespace
        Parsed arguments

    Returns
    -------
    argparse.Namespace or None
        The arguments parsed, updated according to some rules. Exit program
        with error message if error occurs with arguments given.
    """
    models_fastme = {"p-distance": "p", "RY-symetric": "Y", "RY": "R",
                     "JC69": "J", "K2P": "K", "F81": "1", "F84": "4",
                     "TN93": "T", "LogDet": "L"}
    models_fasttree = {"GTR": "-gtr", "JC": ""}
    models_iqtree = set(["HKY", "JC", "F81", "K2P", "K3P", "K81uf",
                         "TNef", "TIM", "TIMef", "TVM", "TVMef", "SYM", "GTR"])
    models_iqtree = {mod: mod for mod in models_iqtree}

    def check_model(models, choice):
        if choice in models.keys():
            return models[choice]
        elif choice in models.values():
            return choice
        mmsg = ("{} is not an available model for {}. Please choose an available DNA model "
                "(see -h for more details)").format(choice, args.soft)
        parser.error(mmsg)

    if args.soft == "quicktree" and args.threads != 1:
        msg = ("You cannot run quicktree with multiple threads. Choose another software, "
               "or remove the --threads option.")
        parser.error(msg)

    if args.soft == "quicktree" and args.model:
        msg = "Quicktree only runs the NJ algorithm. You cannot choose a DNA substitution model."
        parser.error(msg)

    # Memory option only available with iqtree
    if args.soft != "iqtree" and args.soft != "iqtree2" and args.memory:
        msg = "'--mem' option is only available for IQtree."
        parser.error(msg)

    # If bootstraps are asked with iqtree, check the number is >= 1000
    if (args.soft == "iqtree" or args.soft == "iqtree2") and args.boot and int(args.boot) < 1000:
        msg = "With IQtree, number of replicates for bootstraps must be >= 1000."
        parser.error(msg)

    # Write bootstrap option only available for fastme and iqtree
    if (args.soft != "iqtree" and args.soft != "iqtree2" and args.soft != "fastme"
        and args.write_boot):
        msg = "'-B' option is only available with FastME and IQtree."
        parser.error(msg)

    # Fast option only available for iqtree
    if (args.fast and ((args.soft != "iqtree" and args.soft != "iqtree2")
        or (args.boot or args.write_boot))):
        msg = ("-fast option is available only for IQtree, and not compatible "
               "with '-B' and '-b' options (bootstraps).")
        parser.error(msg)

    # Check model name is valid for the chosen soft
    if args.soft == "fastme":
        if args.model:
            args.model = check_model(models_fastme, args.model)
        else:
            args.model = "T"
    elif args.soft == "fasttree":
        if args.model:
            args.model = check_model(models_fasttree, args.model)
        else:
            args.model = "-gtr"
    elif args.soft == "iqtree" or args.soft == "iqtree2":
        if args.model:
            args.model = check_model(models_iqtree, args.model)
        else:
            args.model = "GTR"
    return args


def parse(parser, argu):
    """
    Parse arguments given to parser

    Parameters
    ----------
    parser : argparse.ArgumentParser
        the parser used
    argu : [str]
        command-line given by user, to parse using parser

    Returns
    -------
    argparse.Namespace
        Parsed arguments
    """
    args = parser.parse_args(argu)
    return check_args(parser, args)


if __name__ == '__main__':
    import argparse

    myparser = argparse.ArgumentParser(description=(("Infer phylogenetic tree based on "
                                                     "core/persistent genome")),
                                       add_help=False)
    build_parser(myparser)
    OPTIONS = parse(myparser, sys.argv[1:])
    main_from_parse(OPTIONS)



#!/usr/bin/env python3
# coding: utf-8

"""
Unit tests for general_format_functions.py
"""

import os
import logging
import shutil
from io import StringIO
import pytest

from PanACoTA.annotate_module import format_prokka as prokkafunc
import PanACoTA.utils as utils
import test.test_unit.utilities_for_tests as tutil

ANNOTEDIR = os.path.join("test", "data", "annotate")
EXP_ANNOTE = os.path.join(ANNOTEDIR, "exp_files")
TEST_ANNOTE = os.path.join(ANNOTEDIR, "test_files")
GENEPATH = os.path.join(ANNOTEDIR, "generated_by_unit-tests")

LOGFILE_BASE = os.path.join(GENEPATH, "logfile")
LOGFILES = [LOGFILE_BASE + ext for ext in [".log", ".log.debug", ".log.details", ".log.err"]]


@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    os.mkdir(GENEPATH)
    print("setup")

    yield
    shutil.rmtree(GENEPATH)
    print("teardown")


def test_tbl_to_lst_not_changed_names(caplog):
    """
    Check that generated lstinfo file is as expected, when the genome name is the same as
    it already was in the genome given to prokka.
    The test tblfile contains the following aspects:
    - gene in D strand (start < end)
    - gene in C strand (start > end)
    - CDS features (some with all info = ECnumber, gene name, product etc. ;
    some with missing info)
    - tRNA type
    - repeat_region type (*2) -> should be ignored in .lst
            * 1 in prokka1 version (start end repeat_region
                                        rpt_family CRISPR
                                        score 7)
            * 1 in prokka2 version (start end CRISPR
                                        note CRISPR with x repeat units
                                        rpt_family CRISPR)
    - contigs with more than 2 genes
    - contig with only 2 genes (both 'b' loc)
    - contig with 1 gene ('b' loc)
    - contig without gene (should be skipped)
    """
    caplog.set_level(logging.DEBUG)
    logger = logging.getLogger("test_prokka")
    tblfile = os.path.join(TEST_ANNOTE, "original_name.fna-prokkaRes", "prokka_out_for_test.tbl")
    lstfile = os.path.join(GENEPATH, "res_test_tbl2lst.lst")
    contigs = {"JGIKIPgffgIJ": "test.0417.00002.0001",
               "toto": "test.0417.00002.0002",
               "other_header": "test.0417.00002.0003",
               "my_contig": "test.0417.00002.0004",
               "bis": "test.0417.00002.0005",
               "ter": "test.0417.00002.0006",
               "contname": "test.0417.00002.0007",
              }
    name = "test.0417.00002"
    gpath = "genome_init"
    assert prokkafunc.tbl2lst(tblfile, lstfile, contigs, name, gpath)
    exp_lst = os.path.join(EXP_ANNOTE, "res_create_lst-prokka.lst")
    assert tutil.compare_order_content(exp_lst, lstfile)


def test_tbl_to_lst_changed_names(caplog):
    """
    Check that generated lstinfo file is as expected, when the genome name is not the same as
    it already was in the genome given to prokka.
    The test tblfile contains the following aspects:
    - gene in D strand (start < end)
    - gene in C strand (start > end)
    - CDS features (some with all info = ECnumber, gene name, product etc. ;
    some with missing info)
    - tRNA type
    - repeat_region type (*2)
    - contigs with more than 2 genes
    - contig with only 2 genes (both 'b' loc)
    - contig with 1 gene ('b' loc)
    - contig without gene (should be skipped)
    """
    caplog.set_level(logging.DEBUG)
    logger = logging.getLogger("test_prokka")
    tblfile = os.path.join(TEST_ANNOTE, "prokka_out_tbl_changed-contnames.tbl")
    lstfile = os.path.join(GENEPATH, "res_test_tbl2lst.lst")
    contigs = {"toto_1": "test.0417.00002.0001",
               "toto_2": "test.0417.00002.0002",
               "toto_3": "test.0417.00002.0003",
               "toto_4": "test.0417.00002.0004",
               "toto_5": "test.0417.00002.0005",
               "toto_6": "test.0417.00002.0006",
               "toto_7": "test.0417.00002.0007",
              }
    name = "test.0417.00002"
    gpath = "path_to_genome"
    assert prokkafunc.tbl2lst(tblfile, lstfile, contigs, name, gpath)
    exp_lst = os.path.join(EXP_ANNOTE, "res_create_lst-prokka.lst")
    assert tutil.compare_order_content(exp_lst, lstfile)


def test_tbl_to_lst_unknown(caplog):
    """
    A contig name in the fna file does not exist -> error message, and all result files
    must be removed for this genome
    """
    caplog.set_level(logging.DEBUG)
    logger = logging.getLogger("test_prokka")
    tblfile = os.path.join(TEST_ANNOTE, "prokka_out_tbl_changed-contnames.tbl")
    lstfile = os.path.join(GENEPATH, "res_test_tbl2lst.lst")
    contigs = {"toto_1": "test.0417.00002.0001",
               "toto_2": "test.0417.00002.0002",
               "toto_3": "test.0417.00002.0003",
               "toto_5": "test.0417.00002.0004",
               "toto_6": "test.0417.00002.0006",
               "toto_7": "test.0417.00002.0007",
              }
    name = "test.0417.00002"
    gpath = "path_to_genome"
    assert not prokkafunc.tbl2lst(tblfile, lstfile, contigs, name, gpath)
    assert ("'toto_4' found in test/data/annotate/test_files/prokka_out_tbl_changed-contnames.tbl "
            "does not exist in path_to_genome") in caplog.text


def test_tbl_to_lst_empty_tbl(caplog):
    """
    The tbl file is empty. Error message saying that the file is empty, and returns False
    """
    caplog.set_level(logging.DEBUG)
    logger = logging.getLogger("test_prokka")
    tblfile = os.path.join(GENEPATH, "prokka_out_empty_tbl.tbl")
    open(tblfile, "w").close()
    lstfile = os.path.join(GENEPATH, "res_test_tbl2lst.lst")
    contigs = {"toto_1": "test.0417.00002.0001",
               "toto_2": "test.0417.00002.0002",
               "toto_3": "test.0417.00002.0003",
               "toto_5": "test.0417.00002.0004",
               "toto_6": "test.0417.00002.0006",
               "toto_7": "test.0417.00002.0007",
              }
    name = "test.0417.00002"
    gpath = "path_to_genome"
    assert not prokkafunc.tbl2lst(tblfile, lstfile, contigs, name, gpath)
    assert ("test/data/annotate/generated_by_unit-tests/"
            "prokka_out_empty_tbl.tbl is empty.") in caplog.text


def test_tbl_to_lst_wrong_tbl(caplog):
    """
    The tbl file is in wrong format.
    Error message saying that it is not the expected format, and returns False
    """
    caplog.set_level(logging.DEBUG)
    logger = logging.getLogger("test_prokka")
    tblfile = os.path.join(GENEPATH, "prokka_out_empty_tbl.tbl")
    with open(tblfile, "w") as tblf:
        tblf.write("wrong format")
    lstfile = os.path.join(GENEPATH, "res_test_tbl2lst.lst")
    contigs = {"toto_1": "test.0417.00002.0001",
               "toto_2": "test.0417.00002.0002",
               "toto_3": "test.0417.00002.0003",
               "toto_5": "test.0417.00002.0004",
               "toto_6": "test.0417.00002.0006",
               "toto_7": "test.0417.00002.0007",
              }
    name = "test.0417.00002"
    gpath = "path_to_genome"
    assert not prokkafunc.tbl2lst(tblfile, lstfile, contigs, name, gpath)
    assert ("Wrong format for test/data/annotate/generated_by_unit-tests/"
            "prokka_out_empty_tbl.") in caplog.text


def test_create_gff(caplog):
    """
    Check generated gff file. Must have all sequences in header (even replicons without gene),
    and then 1 line per gene
    """
    caplog.set_level(logging.DEBUG)
    logger = logging.getLogger("test_prodigal")
    gfffile = os.path.join(TEST_ANNOTE, "original_name.fna-prokkaRes",
                           "prokka_out_for_test.gff")
    contigs = {"JGIKIPgffgIJ": "test.0417.00002.0001",
               "toto": "test.0417.00002.0002",
               "other_header": "test.0417.00002.0003",
               "my_contig": "test.0417.00002.0004",
               "bis": "test.0417.00002.0005",
               "ter": "test.0417.00002.0006",
               "contname": "test.0417.00002.0007",
              }
    sizes = {"test.0417.00002.0001": 84,
             "test.0417.00002.0002": 103,
             "test.0417.00002.0003": 122,
             "test.0417.00002.0004": 35,
             "test.0417.00002.0005": 198,
             "test.0417.00002.0006": 128,
             "test.0417.00002.0007": 85
            }
    res_gff_file = os.path.join(GENEPATH, "prodigal_res.gff")
    res_lst = os.path.join(EXP_ANNOTE, "res_create_lst-prokka.lst")
    gpath = "original_genome_name"
    assert prokkafunc.generate_gff(gpath, gfffile, res_gff_file, res_lst, sizes, contigs)
    exp_gff = os.path.join(EXP_ANNOTE, "res_create_gff-prokka.gff")
    assert tutil.compare_order_content(exp_gff, res_gff_file)


def test_create_gff_wrong_start(caplog):
    """
    Check generated gff file. A sequence has not the same start in gff and in tbl
    """
    caplog.set_level(logging.DEBUG)
    logger = logging.getLogger("test_prodigal")
    gfffile = os.path.join(TEST_ANNOTE, "prokka_out_gff-wrong-start.gff")
    contigs = {"JGIKIPgffgIJ": "test.0417.00002.0001",
               "toto": "test.0417.00002.0002",
               "other_header": "test.0417.00002.0003",
               "my_contig": "test.0417.00002.0004",
               "bis": "test.0417.00002.0005",
               "ter": "test.0417.00002.0006",
               "contname": "test.0417.00002.0007",
              }
    sizes = {"test.0417.00002.0001": 84,
             "test.0417.00002.0002": 103,
             "test.0417.00002.0003": 122,
             "test.0417.00002.0004": 35,
             "test.0417.00002.0005": 198,
             "test.0417.00002.0006": 128,
             "test.0417.00002.0007": 85
            }
    res_gff_file = os.path.join(GENEPATH, "prodigal_res.gff")
    res_lst = os.path.join(EXP_ANNOTE, "res_create_lst-prokka.lst")
    gpath = "original_genome_name"
    assert not prokkafunc.generate_gff(gpath, gfffile, res_gff_file, res_lst, sizes, contigs)
    assert ("Files prokka_out_gff-wrong-start.tbl and prokka_out_gff-wrong-start.gff "
            "(in prokka tmp_files: original_genome_name-prokkaRes) do not have the same "
            "start value for gene EPKOMDHM_00011 (3000 in gff, 3500 in tbl)") in caplog.text


def test_create_gff_error_gff(caplog):
    """
    Check generated gff file. The prokka output gff file has a problem (locus_tag != ID)
    -> returns False with error message
    """
    caplog.set_level(logging.DEBUG)
    logger = logging.getLogger("test_prodigal")
    gfffile = os.path.join(TEST_ANNOTE, "prokka_out_gff-error.gff")
    contigs = {"JGIKIPgffgIJ": "test.0417.00002.0001",
               "toto": "test.0417.00002.0002",
               "other_header": "test.0417.00002.0003",
               "my_contig": "test.0417.00002.0004",
               "bis": "test.0417.00002.0005",
               "ter": "test.0417.00002.0006",
               "contname": "test.0417.00002.0007",
              }
    sizes = {"test.0417.00002.0001": 84,
             "test.0417.00002.0002": 103,
             "test.0417.00002.0003": 122,
             "test.0417.00002.0004": 35,
             "test.0417.00002.0005": 198,
             "test.0417.00002.0006": 128,
             "test.0417.00002.0007": 85
            }
    res_gff_file = os.path.join(GENEPATH, "prodigal_res.gff")
    res_lst = os.path.join(EXP_ANNOTE, "res_create_lst-prokka.lst")
    gpath = "original_genome_name"
    assert not prokkafunc.generate_gff(gpath, gfffile, res_gff_file, res_lst, sizes, contigs)
    assert ("Problem in prokka_out_gff-error.gff: ID=EPKOMDHM_00005 whereas "
            "locus_tag=toto") in caplog.text


def test_create_gff_wrong_format(caplog):
    """
    Check generated gff file. The prokka output gff file has a problem (locus_tag != ID)
    -> returns False with error message
    """
    caplog.set_level(logging.DEBUG)
    logger = logging.getLogger("test_prodigal")
    gfffile = os.path.join(GENEPATH, "prokka_out_wrong_format.gff")
    with open(gfffile, "w") as gfff:
        gfff.write("##gff-version3\n")
        gfff.write("##sequence-region bis 1 600\n")
        gfff.write("JGIKIPgffgIJ    Prodigal:2.6    CDS 287 787\n")
    contigs = {"JGIKIPgffgIJ": "test.0417.00002.0001",
               "toto": "test.0417.00002.0002",
               "other_header": "test.0417.00002.0003",
               "my_contig": "test.0417.00002.0004",
               "bis": "test.0417.00002.0005",
               "ter": "test.0417.00002.0006",
               "contname": "test.0417.00002.0007",
              }
    sizes = {"test.0417.00002.0001": 84,
             "test.0417.00002.0002": 103,
             "test.0417.00002.0003": 122,
             "test.0417.00002.0004": 35,
             "test.0417.00002.0005": 198,
             "test.0417.00002.0006": 128,
             "test.0417.00002.0007": 85
            }
    res_gff_file = os.path.join(GENEPATH, "prodigal_res.gff")
    res_lst = os.path.join(EXP_ANNOTE, "res_create_lst-prokka.lst")
    gpath = "original_genome_name"
    assert not prokkafunc.generate_gff(gpath, gfffile, res_gff_file, res_lst, sizes, contigs)
    assert ("Wrong format for test/data/annotate/generated_by_unit-tests/"
            "prokka_out_wrong_format.gff") in caplog.text


def test_create_gen(caplog):
    """
    Check create gen file.
    """
    caplog.set_level(logging.DEBUG)
    logger = logging.getLogger("test_prodigal")
    ffnfile = os.path.join(TEST_ANNOTE, "original_name.fna-prokkaRes",
                           "prokka_out_for_test.ffn")
    lstfile = os.path.join(EXP_ANNOTE, "res_create_lst-prokka.lst")
    res_gen_file = os.path.join(GENEPATH, "prodigal_res.gen")
    assert prokkafunc.create_gen(ffnfile, lstfile, res_gen_file)
    exp_gen = os.path.join(EXP_ANNOTE, "res_create_gene_prokka.gen")
    assert tutil.compare_order_content(exp_gen, res_gen_file)


def test_create_gen_supgen(caplog):
    """
    Check create gen file. But there is a gene in ffn that is not in lst -> error
    """
    caplog.set_level(logging.DEBUG)
    logger = logging.getLogger("test_prodigal")
    ffnfile = os.path.join(TEST_ANNOTE, "prokka_out_for_test-supGene.ffn")
    lstfile = os.path.join(EXP_ANNOTE, "res_create_lst-prokka.lst")
    res_gen_file = os.path.join(GENEPATH, "prodigal_res.gen")
    assert not prokkafunc.create_gen(ffnfile, lstfile, res_gen_file)
    assert ("Missing info for gene >JGIKIPIJ_03050 (from test/data/annotate/test_files/prokka_out_for_test-supGene.ffn) in test/data/annotate/exp_files/res_create_lst-prokka.lst. If it is actually present in the lst file, check that genes are ordered by increasing number in both lst and ffn files.") in caplog.text


def test_create_gen_missingSeq(caplog):
    """
    Check create gen file. A gene in lst does not have a sequence in ffn.
    Just skip it, and go to next sequence for gen file.
    """
    caplog.set_level(logging.DEBUG)
    logger = logging.getLogger("test_prodigal")
    ffnfile = os.path.join(TEST_ANNOTE, "prokka_out_for_test-noSeqFor1gene.ffn")
    lstfile = os.path.join(EXP_ANNOTE, "res_create_lst-prokka.lst")
    res_gen_file = os.path.join(GENEPATH, "prodigal_res.gen")
    assert prokkafunc.create_gen(ffnfile, lstfile, res_gen_file)
    exp_gen = os.path.join(EXP_ANNOTE, "res_create_gene_prokka-missGene.gen")
    assert tutil.compare_order_content(exp_gen, res_gen_file)


def test_create_prt(caplog):
    """
    Check that prt file is generated as expected
    """
    caplog.set_level(logging.DEBUG)
    protfile = os.path.join(TEST_ANNOTE, "original_name.fna-prokkaRes",
                           "prokka_out_for_test.faa")
    res_prt_file = os.path.join(GENEPATH, "prokka_res.prt")
    lstfile = os.path.join(EXP_ANNOTE, "res_create_lst-prokka.lst")
    exp_lst = os.path.join(EXP_ANNOTE, "res_create_gene_lst_prodigal.lst")
    assert prokkafunc.create_prt(protfile, lstfile, res_prt_file)
    exp_prt = os.path.join(EXP_ANNOTE, "res_create_prt_prokka.faa")
    assert tutil.compare_order_content(exp_prt, res_prt_file)


def test_create_prt_wrong_header_int(caplog):
    """
    Test creating prt file, but the faa file has a header with wrong format (>JGIKIPIJ_d0008)
    """
    caplog.set_level(logging.DEBUG)
    protfile = os.path.join(TEST_ANNOTE, "prokka_out_for_test-wrongHeaderInt.faa")
    res_prt_file = os.path.join(GENEPATH, "prokka_res.prt")
    lstfile = os.path.join(EXP_ANNOTE, "res_create_lst-prokka.lst")
    exp_lst = os.path.join(EXP_ANNOTE, "res_create_gene_lst_prodigal.lst")
    assert not prokkafunc.create_prt(protfile, lstfile, res_prt_file)
    assert ("Unknown header format >JGIKIPIJ_d0008 in test/data/annotate/test_files/"
            "prokka_out_for_test-wrongHeaderInt.faa. Gene ID is not a number.") in caplog.text


def test_create_prt_wrong_header_sep(caplog):
    """
    Test creating prt file, but the faa file has a header with wrong format,
    no '_' between name and gene ID (>JGIKIPIJ00008)
    """
    caplog.set_level(logging.DEBUG)
    protfile = os.path.join(TEST_ANNOTE, "prokka_out_for_test-wrongHeaderSep.faa")
    res_prt_file = os.path.join(GENEPATH, "prokka_res.prt")
    lstfile = os.path.join(EXP_ANNOTE, "res_create_lst-prokka.lst")
    exp_lst = os.path.join(EXP_ANNOTE, "res_create_gene_lst_prodigal.lst")
    assert not prokkafunc.create_prt(protfile, lstfile, res_prt_file)
    assert ("Unknown header format >JGIKIPIJ00008 in test/data/annotate/test_files/"
            "prokka_out_for_test-wrongHeaderSep.faa. Gene ID is not a number.") in caplog.text


def test_create_prt_wrong_unknown_prot(caplog):
    """
    Test creating prt file, but the faa file has a protein (>sup-prot_00012)
    which is not in the lst file
    """
    caplog.set_level(logging.DEBUG)
    protfile = os.path.join(TEST_ANNOTE, "prokka_out_for_test-supHeader.faa")
    res_prt_file = os.path.join(GENEPATH, "prokka_res.prt")
    lstfile = os.path.join(EXP_ANNOTE, "res_create_lst-prokka.lst")
    exp_lst = os.path.join(EXP_ANNOTE, "res_create_gene_lst_prodigal.lst")
    assert not prokkafunc.create_prt(protfile, lstfile, res_prt_file)
    assert ("Missing info for protein >sup-prot_00012 (from test/data/annotate/test_files/"
            "prokka_out_for_test-supHeader.faa) in test/data/annotate/exp_files/"
            "res_create_lst-prokka.lst. If it is actually present in the lst file, check that "
            "proteins are ordered by increasing number in both lst and faa files.") in caplog.text


def test_format_1genome(caplog):
    """
    Test that when prokka results are ok, all files are generated as expected.
    """
    caplog.set_level(logging.DEBUG)
    name = "test.0417.00002"
    # path to original genome, given to prodigal for annotation
    gpath =  os.path.join(TEST_ANNOTE, "original_name.fna")
    prok_path = TEST_ANNOTE
    # Create result directories
    prot_dir = os.path.join(GENEPATH, "Proteins")
    lst_dir = os.path.join(GENEPATH, "LSTINFO")
    rep_dir = os.path.join(GENEPATH, "Replicons")
    gene_dir = os.path.join(GENEPATH, "Genes")
    gff_dir = os.path.join(GENEPATH, "gff")
    os.makedirs(prot_dir)
    os.makedirs(lst_dir)
    os.makedirs(rep_dir)
    os.makedirs(gene_dir)
    os.makedirs(gff_dir)

    assert prokkafunc.format_one_genome(gpath, name, prok_path, lst_dir, prot_dir, gene_dir,
                                        rep_dir, gff_dir)

    # Check output files content
    # Replicons
    exp_rep = os.path.join(EXP_ANNOTE, "res_created_rep-prokka.fna")
    res_rep_file = os.path.join(rep_dir, "test.0417.00002.fna")
    assert tutil.compare_order_content(exp_rep, res_rep_file)
    # Proteins
    exp_prt = os.path.join(EXP_ANNOTE, "res_create_prt_prokka.faa")
    res_prt_file = os.path.join(prot_dir, "test.0417.00002.prt")
    assert tutil.compare_order_content(exp_prt, res_prt_file)
    # Genes
    exp_gen = os.path.join(EXP_ANNOTE, "res_create_gene_prokka.gen")
    res_gen_file = os.path.join(gene_dir, "test.0417.00002.gen")
    assert tutil.compare_order_content(exp_gen, res_gen_file)
    # LSTINFO
    exp_lst = os.path.join(EXP_ANNOTE, "res_create_lst-prokka.lst")
    res_lst_file = os.path.join(lst_dir, "test.0417.00002.lst")
    assert tutil.compare_order_content(exp_lst, res_lst_file)
    # gff
    exp_gff = os.path.join(EXP_ANNOTE, "res_create_gff-prokka.gff")
    res_gff_file = os.path.join(gff_dir, "test.0417.00002.gff")
    assert tutil.compare_order_content(exp_gff, res_gff_file)


def test_format_1genome_emptygpath(caplog):
    """
    Test on formatting prokka results, when original ffn file is empty -> error message,
    and no file generated
    """
    caplog.set_level(logging.DEBUG)
    name = "prokka_out_for_test"
    # Create empty file, that we give to prodigal for formatting step
    gpath =  os.path.join(GENEPATH, "original_name-empty.fna")
    open(gpath, "w").close()
    # Create prokka result files (empty files, will not be read)
    gpath_prokres =  gpath + "-prokkaRes"
    os.makedirs(gpath_prokres)
    fna_prokres = os.path.join(gpath_prokres, "prokka_out_for_test.fna")
    open(fna_prokres, "w").close()
    tbl_prokres = os.path.join(gpath_prokres, "prokka_out_for_test.tbl")
    open(tbl_prokres, "w").close()
    gff_prokres = os.path.join(gpath_prokres, "prokka_out_for_test.gff")
    open(gff_prokres, "w").close()
    ffn_prokres = os.path.join(gpath_prokres, "prokka_out_for_test.ffn")
    open(ffn_prokres, "w").close()
    faa_prokres = os.path.join(gpath_prokres, "prokka_out_for_test.faa")
    open(faa_prokres, "w").close()
    # Create result directories
    prok_path = GENEPATH
    prot_dir = os.path.join(GENEPATH, "Proteins")
    lst_dir = os.path.join(GENEPATH, "LSTINFO")
    rep_dir = os.path.join(GENEPATH, "Replicons")
    gen_dir = os.path.join(GENEPATH, "Genes")
    gff_dir = os.path.join(GENEPATH, "gff")
    os.makedirs(rep_dir)
    os.makedirs(gff_dir)
    os.makedirs(lst_dir)
    os.makedirs(gen_dir)
    # Add empty res lst, gff and gen files, to check that it is removed at the end
    res_gff_file = os.path.join(gff_dir, "prokka_out_for_test.gff")
    open(res_gff_file, "w").close()
    assert len(os.listdir(gff_dir) ) == 1
    res_lst_file = os.path.join(lst_dir, "prokka_out_for_test.lst")
    open(res_lst_file, "w").close()
    assert len(os.listdir(lst_dir) ) == 1
    res_gen_file = os.path.join(gen_dir, "prokka_out_for_test.gen")
    open(res_gen_file, "w").close()
    assert len(os.listdir(gen_dir) ) == 1
    # res_gen_file = os.path.join(gen_dir, "prokka_out_for_test.gen")
    # open(res_gen_file, "w").close()
    # assert len(os.listdir(gen_dir) ) == 1

    assert not prokkafunc.format_one_genome(gpath, name, prok_path, lst_dir, prot_dir, gen_dir,
                                              rep_dir, gff_dir)
    # Check that all files were removed
    assert len(os.listdir(rep_dir) ) == 0
    assert len(os.listdir(lst_dir) ) == 0
    assert len(os.listdir(gff_dir) ) == 0
    assert len(os.listdir(gen_dir) ) == 0
    # Check log
    assert ("Problems while generating Replicon file for prokka_out_for_test") in caplog.text


def test_format_1genome_pb_tbl(caplog):
    """
    Test on formatting prokka results, when prokka output tbl file does not have
    the expected format -> error message, and no file generated
    """
    caplog.set_level(logging.DEBUG)
    name = "test.0417.00002"
    # path to original genome, given to prodigal for annotation
    orig_gpath =  os.path.join(TEST_ANNOTE, "original_name.fna")
    # In generated_by_tests folder, create the original genome given to prokka
    # (copy from test_file)
    used_gpath = os.path.join(GENEPATH, "original_name.fna")
    used_respath = used_gpath + "-prokkaRes"
    os.makedirs(used_respath)
    shutil.copyfile(orig_gpath, used_gpath)

    # Create tbl_file with a wrong format
    with open(os.path.join(used_respath, "prokka_out_for_test.tbl"), "w") as ori:
        ori.write(">wrongheader # 1 # 2 # 1 # toto")
    # Add empty prokka res gff ffn and faa files (they won't be read, as it will stop
    # at tbl2lst)
    orig_fna =  os.path.join(TEST_ANNOTE, "original_name.fna-prokkaRes", "prokka_out_for_test.fna")
    fna_prokres = os.path.join(used_respath, "prokka_out_for_test.fna")
    shutil.copyfile(orig_fna, fna_prokres)
    res_gff_file = os.path.join(used_respath, "prokka_out_for_test.gff")
    open(res_gff_file, "w").close()
    res_ffn_file = os.path.join(used_respath, "prokka_out_for_test.ffn")
    open(res_ffn_file, "w").close()
    res_faa_file = os.path.join(used_respath, "prokka_out_for_test.faa")
    open(res_faa_file, "w").close()

    # Create output directories
    prok_path = GENEPATH
    prot_dir = os.path.join(GENEPATH, "Proteins")
    lst_dir = os.path.join(GENEPATH, "LSTINFO")
    rep_dir = os.path.join(GENEPATH, "Replicons")
    gen_dir = os.path.join(GENEPATH, "Genes")
    gff_dir = os.path.join(GENEPATH, "gff")
    os.makedirs(rep_dir)
    os.makedirs(gff_dir)
    os.makedirs(lst_dir)
    os.makedirs(gen_dir)
    # Add empty res lst, gff and gen files, to check that it is removed at the end
    res_gff_file = os.path.join(gff_dir, "test.0417.00002.gff")
    open(res_gff_file, "w").close()
    assert len(os.listdir(gff_dir) ) == 1
    res_lst_file = os.path.join(lst_dir, "test.0417.00002.lst")
    open(res_lst_file, "w").close()
    assert len(os.listdir(lst_dir) ) == 1
    res_gen_file = os.path.join(gen_dir, "test.0417.00002.gen")
    open(res_gen_file, "w").close()
    assert len(os.listdir(gen_dir) ) == 1

    # Run formatting
    assert not prokkafunc.format_one_genome(used_gpath, name, prok_path, lst_dir, prot_dir,
                                            gen_dir, rep_dir, gff_dir)

    # Check that all files were removed
    assert len(os.listdir(rep_dir) ) == 0
    assert len(os.listdir(lst_dir) ) == 0
    assert len(os.listdir(gff_dir) ) == 0
    assert len(os.listdir(gen_dir) ) == 0
    assert("Wrong format for test/data/annotate/generated_by_unit-tests/"
           "original_name.fna-prokkaRes/prokka_out_for_test.tbl.") in caplog.text
    assert ("Problems while generating LSTINFO file for test.0417.00002") in caplog.text


def test_format_1genome_pb_gff(caplog):
    """
    Test on formatting prokka results, when prokka output gff file does not have
    the expected format -> error message, and no file generated
    """
    caplog.set_level(logging.DEBUG)
    name = "test.0417.00002"
    # path to original genome, given to prodigal for annotation
    orig_gpath =  os.path.join(TEST_ANNOTE, "original_name.fna")
    # In generated_by_tests folder, create the original genome given to prokka
    # (copy from test_file)
    used_gpath = os.path.join(GENEPATH, "original_name.fna")
    used_respath = used_gpath + "-prokkaRes"
    os.makedirs(used_respath)
    shutil.copyfile(orig_gpath, used_gpath)
    # Copy tbl file, which is as expected (tbl2lst must succeed)
    orig_tbl = os.path.join(orig_gpath + "-prokkaRes",
                            "prokka_out_for_test.tbl")
    used_tbl = os.path.join(used_respath, "prokka_out_for_test.tbl")
    shutil.copyfile(orig_tbl, used_tbl)
    orig_fna =  os.path.join(orig_gpath + "-prokkaRes", "prokka_out_for_test.fna")
    fna_prokres = os.path.join(used_respath, "prokka_out_for_test.fna")
    shutil.copyfile(orig_fna, fna_prokres)

    # Create gff_file with a wrong format
    with open(os.path.join(used_respath, "prokka_out_for_test.gff"), "w") as ori:
        ori.write("wrongheader # 1 # 2 # 1 # toto")
    # Add empty prokka res ffn and faa files
    res_ffn_file = os.path.join(used_respath, "prokka_out_for_test.ffn")
    open(res_ffn_file, "w").close()
    res_faa_file = os.path.join(used_respath, "prokka_out_for_test.faa")
    open(res_faa_file, "w").close()

    # Create output directories
    prok_path = GENEPATH
    prot_dir = os.path.join(GENEPATH, "Proteins")
    lst_dir = os.path.join(GENEPATH, "LSTINFO")
    rep_dir = os.path.join(GENEPATH, "Replicons")
    gen_dir = os.path.join(GENEPATH, "Genes")
    gff_dir = os.path.join(GENEPATH, "gff")
    os.makedirs(rep_dir)
    os.makedirs(gff_dir)
    os.makedirs(lst_dir)
    os.makedirs(gen_dir)
    # Add empty res lst, gff and gen files, to check that it is removed at the end
    res_gff_file = os.path.join(gff_dir, "test.0417.00002.gff")
    open(res_gff_file, "w").close()
    assert len(os.listdir(gff_dir) ) == 1
    res_lst_file = os.path.join(lst_dir, "test.0417.00002.lst")
    open(res_lst_file, "w").close()
    assert len(os.listdir(lst_dir) ) == 1
    res_gen_file = os.path.join(gen_dir, "test.0417.00002.gen")
    open(res_gen_file, "w").close()
    assert len(os.listdir(gen_dir) ) == 1

    # Run formatting
    assert not prokkafunc.format_one_genome(used_gpath, name, prok_path, lst_dir, prot_dir,
                                            gen_dir, rep_dir, gff_dir)

    # Check that all files were removed
    assert len(os.listdir(rep_dir) ) == 0
    assert len(os.listdir(lst_dir) ) == 0
    assert len(os.listdir(gff_dir) ) == 0
    assert len(os.listdir(gen_dir) ) == 0
    assert("Wrong format for test/data/annotate/generated_by_unit-tests/"
           "original_name.fna-prokkaRes/prokka_out_for_test.gff.") in caplog.text
    assert ("Problems while generating .gff file for test.0417.00002") in caplog.text


def test_format_1genome_pb_ffn(caplog):
    """
    Test on formatting prokka results, when prokka output ffn file does not have
    the expected format -> error message, and no file generated
    """
    caplog.set_level(logging.DEBUG)
    name = "test.0417.00002"
    # path to original genome, given to prodigal for annotation
    orig_gpath =  os.path.join(TEST_ANNOTE, "original_name.fna")
    # In generated_by_tests folder, create the original genome given to prokka
    # (copy from test_file)
    used_gpath = os.path.join(GENEPATH, "original_name.fna")
    used_respath = used_gpath + "-prokkaRes"
    os.makedirs(used_respath)
    shutil.copyfile(orig_gpath, used_gpath)
    # Copy tbl and gff files, which is as expected (tbl2lst and generate_gff must succeed)
    orig_tbl = os.path.join(orig_gpath + "-prokkaRes",
                            "prokka_out_for_test.tbl")
    used_tbl = os.path.join(used_respath, "prokka_out_for_test.tbl")
    shutil.copyfile(orig_tbl, used_tbl)
    orig_gff = os.path.join(orig_gpath + "-prokkaRes",
                            "prokka_out_for_test.gff")
    used_gff = os.path.join(used_respath, "prokka_out_for_test.gff")
    shutil.copyfile(orig_gff, used_gff)
    orig_fna =  os.path.join(orig_gpath + "-prokkaRes", "prokka_out_for_test.fna")
    fna_prokres = os.path.join(used_respath, "prokka_out_for_test.fna")
    shutil.copyfile(orig_fna, fna_prokres)

    # Create ffn_file with a wrong format
    orig_ffn = os.path.join(TEST_ANNOTE, "prokka_out_for_test-supGene.ffn")
    used_ffn = os.path.join(used_respath, "prokka_out_for_test.ffn")
    shutil.copyfile(orig_ffn, used_ffn)
    # Add empty prokka res faa file
    res_faa_file = os.path.join(used_respath, "prokka_out_for_test.faa")
    open(res_faa_file, "w").close()

    # Create output directories
    prok_path = GENEPATH
    prot_dir = os.path.join(GENEPATH, "Proteins")
    lst_dir = os.path.join(GENEPATH, "LSTINFO")
    rep_dir = os.path.join(GENEPATH, "Replicons")
    gen_dir = os.path.join(GENEPATH, "Genes")
    gff_dir = os.path.join(GENEPATH, "gff")
    os.makedirs(rep_dir)
    os.makedirs(gff_dir)
    os.makedirs(lst_dir)
    os.makedirs(gen_dir)
    # Add empty res lst, gff and gen files, to check that it is removed at the end
    res_gff_file = os.path.join(gff_dir, "test.0417.00002.gff")
    open(res_gff_file, "w").close()
    assert len(os.listdir(gff_dir) ) == 1
    res_lst_file = os.path.join(lst_dir, "test.0417.00002.lst")
    open(res_lst_file, "w").close()
    assert len(os.listdir(lst_dir) ) == 1
    res_gen_file = os.path.join(gen_dir, "test.0417.00002.gen")
    open(res_gen_file, "w").close()
    assert len(os.listdir(gen_dir) ) == 1

    # Run formatting
    assert not prokkafunc.format_one_genome(used_gpath, name, prok_path, lst_dir, prot_dir,
                                            gen_dir, rep_dir, gff_dir)

    # Check that all files were removed
    assert len(os.listdir(rep_dir) ) == 0
    assert len(os.listdir(lst_dir) ) == 0
    assert len(os.listdir(gff_dir) ) == 0
    assert len(os.listdir(gen_dir) ) == 0
    assert("Missing info for gene >JGIKIPIJ_03050 (from test/data/annotate/"
           "generated_by_unit-tests/original_name.fna-prokkaRes/prokka_out_for_test.ffn) "
           "in test/data/annotate/generated_by_unit-tests/LSTINFO/test.0417.00002.lst. "
           "If it is actually present in the lst file, "
           "check that genes are ordered by increasing number in both lst and "
           "ffn files.") in caplog.text
    assert ("Problems while generating .gen file for test.0417.00002") in caplog.text


def test_format_1genome_pb_faa(caplog):
    """
    Test on formatting prokka results, when prokka output faa file does not have
    the expected format -> error message, and no file generated
    """
    caplog.set_level(logging.DEBUG)
    name = "test.0417.00002"
    # path to original genome, given to prodigal for annotation
    orig_gpath =  os.path.join(TEST_ANNOTE, "original_name.fna")
    # In generated_by_tests folder, create the original genome given to prokka
    # (copy from test_file)
    used_gpath = os.path.join(GENEPATH, "original_name.fna")
    used_respath = used_gpath + "-prokkaRes"
    os.makedirs(used_respath)
    shutil.copyfile(orig_gpath, used_gpath)
    # Copy tbl and gff files, which is as expected (tbl2lst and generate_gff must succeed)
    orig_fna =  os.path.join(orig_gpath + "-prokkaRes", "prokka_out_for_test.fna")
    fna_prokres = os.path.join(used_respath, "prokka_out_for_test.fna")
    shutil.copyfile(orig_fna, fna_prokres)
    orig_tbl = os.path.join(orig_gpath + "-prokkaRes",
                            "prokka_out_for_test.tbl")
    used_tbl = os.path.join(used_respath, "prokka_out_for_test.tbl")
    shutil.copyfile(orig_tbl, used_tbl)
    orig_gff = os.path.join(orig_gpath + "-prokkaRes",
                            "prokka_out_for_test.gff")
    used_gff = os.path.join(used_respath, "prokka_out_for_test.gff")
    shutil.copyfile(orig_gff, used_gff)
    orig_ffn = os.path.join(orig_gpath + "-prokkaRes",
                            "prokka_out_for_test.ffn")
    used_ffn = os.path.join(used_respath, "prokka_out_for_test.ffn")
    shutil.copyfile(orig_ffn, used_ffn)
    # Create faa_file with a wrong format
    orig_faa = os.path.join(TEST_ANNOTE, "prokka_out_for_test-wrongHeaderInt.faa")
    used_faa = os.path.join(used_respath, "prokka_out_for_test.faa")
    shutil.copyfile(orig_faa, used_faa)

    # Create output directories
    prok_path = GENEPATH
    prot_dir = os.path.join(GENEPATH, "Proteins")
    lst_dir = os.path.join(GENEPATH, "LSTINFO")
    rep_dir = os.path.join(GENEPATH, "Replicons")
    gen_dir = os.path.join(GENEPATH, "Genes")
    gff_dir = os.path.join(GENEPATH, "gff")
    os.makedirs(rep_dir)
    os.makedirs(gff_dir)
    os.makedirs(lst_dir)
    os.makedirs(gen_dir)
    os.makedirs(prot_dir)
    # Add empty res lst, gff and gen files, to check that it is removed at the end
    res_gff_file = os.path.join(gff_dir, "test.0417.00002.gff")
    open(res_gff_file, "w").close()
    assert len(os.listdir(gff_dir) ) == 1
    res_lst_file = os.path.join(lst_dir, "test.0417.00002.lst")
    open(res_lst_file, "w").close()
    assert len(os.listdir(lst_dir) ) == 1
    res_gen_file = os.path.join(gen_dir, "test.0417.00002.gen")
    open(res_gen_file, "w").close()
    assert len(os.listdir(gen_dir) ) == 1

    # Run formatting
    assert not prokkafunc.format_one_genome(used_gpath, name, prok_path, lst_dir, prot_dir,
                                            gen_dir, rep_dir, gff_dir)

    # Check that all files were removed
    assert len(os.listdir(rep_dir) ) == 0
    assert len(os.listdir(lst_dir) ) == 0
    assert len(os.listdir(gff_dir) ) == 0
    assert len(os.listdir(gen_dir) ) == 0
    assert("Unknown header format >JGIKIPIJ_d0008 in test/data/annotate/generated_by_unit-tests/"
           "original_name.fna-prokkaRes/prokka_out_for_test.faa. "
           "Gene ID is not a number.") in caplog.text
    assert ("Problems while generating .prt file for test.0417.00002") in caplog.text



#!/usr/bin/env python3
# coding: utf-8

"""
Functions used by tree unit tests
"""
import numbers

from Bio import Phylo


def is_tree_lengths(treefile):
    """
    Check that given tree is in newick format with branch lengths only (no bootstrap)

    Parameters
    ----------
    treefile: str
        Path to file containing tree to check

    Returns
    -------
    bool
    """
    mytree = Phylo.read(treefile, "newick")
    root = False
    for elem in mytree.find_clades():
        # Check that each clade has a branch length and no bootstrap
        if not isinstance(elem.branch_length, float):
            if not root:
                root = True
            else:
                print("No branch length")
                return False
        if elem.confidence is not None:
            print("Bootstrap value")
            return False
    return True


def is_tree_bootstrap(treefile):
        """
        Check that given tree is in newick format with branch lengths and bootstrap values

        Parameters
        ----------
        treefile: str
            Path to file containing tree to check

        Returns
        -------
        bool
        """
        mytree = Phylo.read(treefile, "newick")
        root = False
        for elem in mytree.find_clades():
            # Check that each clade has a branch length
            if not isinstance(elem.branch_length, numbers.Real):
                if not root:
                    root = True
                    continue
                else:
                    print("No branch length")
                    return False
            # If it is a leaf, check that no bootstrap value. Otherwise, check that bootstrap value
            if elem.name is None:
                if not isinstance(elem.confidence, numbers.Real):
                    print("No bootstrap", elem.confidence)
                    return False
            else:
                if elem.confidence is not None:
                    print("Bootstrap for leaf")
                    return False
        return True



#!/usr/bin/env python3
# coding: utf-8

"""
Unit tests for the mmseqs_functions submodule in pangenome module
"""

import os
import time
import shutil
import glob
import logging
import pytest

import PanACoTA.pangenome_module.mmseqs_functions as mmseqs
import PanACoTA.utils as utils
import test.test_unit.utilities_for_tests as tutil

LOGFILE_BASE = "logfile_test.txt"
LEVEL = logging.DEBUG
LOGFILES = [LOGFILE_BASE + ext for ext in [".log", ".log.debug", ".log.details", ".log.err"]]
# Define variables shared by several tests
PANDIR = os.path.join("test", "data", "pangenome")
PATH_TEST_FILES = os.path.join(PANDIR, "test_files")
PATH_EXP_FILES = os.path.join(PANDIR, "exp_files")
GENEPATH = os.path.join(PANDIR, "generated_by_unit-tests")


@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    utils.init_logger(LOGFILE_BASE, logging.DEBUG, 'test_getseq', verbose=1)
    os.mkdir(GENEPATH)
    print("setup")

    yield
    shutil.rmtree(GENEPATH)
    for f in LOGFILES:
        if os.path.exists(f):
            os.remove(f)
    print("teardown")

# Clusters expected for the given databank (ref: [members]
EXP_CLUSTERS = {"GEN2.1017.00001.i0002_00004": ["GEN2.1017.00001.i0002_00004",
                                                "GEN4.1111.00001.i0001_00002",
                                                "GENO.1017.00001.b0002_00003",
                                                "GENO.1216.00002.i0001_00003"],
                "GEN2.1017.00001.b0003_00010": ["GEN2.1017.00001.b0003_00010"],
                "GEN2.1017.00001.b0004_00013": ["GEN2.1017.00001.b0004_00013"],
                "GEN4.1111.00001.b0001_00001": ["GEN2.1017.00001.i0002_00005",
                                                "GEN4.1111.00001.b0001_00001",
                                                "GENO.1017.00001.b0001_00002",
                                                "GENO.1216.00002.b0001_00001",
                                                "GENO.1216.00002.i0001_00002"],
                "GEN4.1111.00001.i0001_00003": ["GEN4.1111.00001.i0001_00003"],
                "GEN4.1111.00001.b0001_00009": ["GEN2.1017.00001.b0004_00011",
                                                "GEN4.1111.00001.b0001_00009",
                                                "GENO.1017.00001.b0002_00011",
                                                "GENO.1216.00002.b0002_00010"],
                "GENO.1017.00001.b0001_00001": ["GEN2.1017.00001.b0002_00006",
                                                "GENO.1017.00001.b0001_00001"],
                "GENO.1017.00001.i0002_00006": ["GEN2.1017.00001.b0003_00007",
                                                "GEN4.1111.00001.i0001_00006",
                                                "GENO.1017.00001.i0002_00006",
                                                "GENO.1017.00001.i0002_00007",
                                                "GENO.1216.00002.i0001_00007"],
                "GENO.1017.00001.i0002_00008": ["GEN2.1017.00001.i0004_00012",
                                                "GENO.1017.00001.i0002_00008"],
                "GENO.1017.00001.i0002_00009": ["GEN2.1017.00001.i0003_00008",
                                                "GEN4.1111.00001.i0001_00007",
                                                "GENO.1017.00001.i0002_00009",
                                                "GENO.1216.00002.b0001_00008"],
                "GENO.1017.00001.i0002_00010": ["GEN2.1017.00001.i0003_00009",
                                                "GEN4.1111.00001.i0001_00008",
                                                "GENO.1017.00001.i0002_00010",
                                                "GENO.1216.00002.b0002_00009"],
                "GENO.1216.00002.i0001_00004": ["GEN2.1017.00001.b0002_00003",
                                                "GENO.1216.00002.i0001_00004"],
                "GENO.1216.00002.i0001_00005": ["GEN2.1017.00001.b0001_00002",
                                                "GEN4.1111.00001.i0001_00004",
                                                "GENO.1017.00001.i0002_00004",
                                                "GENO.1216.00002.i0001_00005"],
                "GENO.1216.00002.i0001_00006": ["GEN2.1017.00001.b0001_00001",
                                                "GEN4.1111.00001.i0001_00005",
                                                "GENO.1017.00001.i0002_00005",
                                                "GENO.1216.00002.i0001_00006"],
                "GENO.1216.00002.b0003_00011": ["GENO.1216.00002.b0003_00011"],
                "GENO.1216.00002.b0003_00012": ["GENO.1216.00002.b0003_00012"]
                }

# protein families expected
FAMILIES4G = [["GEN2.1017.00001.i0002_00004", "GEN4.1111.00001.i0001_00002",
               "GENO.1017.00001.b0002_00003", "GENO.1216.00002.i0001_00003"],
              ["GEN2.1017.00001.b0003_00010"],
              ["GEN2.1017.00001.b0004_00013"],
              ["GEN2.1017.00001.i0002_00005", "GEN4.1111.00001.b0001_00001",
               "GENO.1017.00001.b0001_00002", "GENO.1216.00002.b0001_00001",
               "GENO.1216.00002.i0001_00002"],
              ["GEN4.1111.00001.i0001_00003"],
              ["GEN2.1017.00001.b0004_00011", "GEN4.1111.00001.b0001_00009",
               "GENO.1017.00001.b0002_00011", "GENO.1216.00002.b0002_00010"],
              ["GEN2.1017.00001.b0002_00006", "GENO.1017.00001.b0001_00001"],
              ["GEN2.1017.00001.b0003_00007", "GEN4.1111.00001.i0001_00006",
               "GENO.1017.00001.i0002_00006", "GENO.1017.00001.i0002_00007",
               "GENO.1216.00002.i0001_00007"],
              ["GEN2.1017.00001.i0004_00012", "GENO.1017.00001.i0002_00008"],
              ["GEN2.1017.00001.i0003_00008", "GEN4.1111.00001.i0001_00007",
               "GENO.1017.00001.i0002_00009", "GENO.1216.00002.b0001_00008"],
              ["GEN2.1017.00001.i0003_00009", "GEN4.1111.00001.i0001_00008",
               "GENO.1017.00001.i0002_00010", "GENO.1216.00002.b0002_00009"],
              ["GEN2.1017.00001.b0002_00003", "GENO.1216.00002.i0001_00004"],
              ["GEN2.1017.00001.b0001_00002", "GEN4.1111.00001.i0001_00004",
               "GENO.1017.00001.i0002_00004", "GENO.1216.00002.i0001_00005"],
              ["GEN2.1017.00001.b0001_00001", "GEN4.1111.00001.i0001_00005",
               "GENO.1017.00001.i0002_00005", "GENO.1216.00002.i0001_00006"],
              ["GENO.1216.00002.b0003_00011"],
              ["GENO.1216.00002.b0003_00012"]
              ]


# Start tests
def test_create_mmseqdb(caplog):
    """
    Test that mmseq DB is created. We do not check its content as it could change
    according to mmseq versions, and we are here testing PanACoTA, not mmseqs
    Just check that all expected outfiles are present.
    """
    caplog.set_level(logging.DEBUG)
    filename = os.path.join(GENEPATH, "test_create_mmseqsdb.msdb")
    prt_path = os.path.join(PATH_EXP_FILES, "exp_EXEM.All.prt")
    logfile = os.path.join(GENEPATH, "test_create_mmseqsdb.log")
    mmseqs.create_mmseqs_db(filename, prt_path, logfile)

    outext = ["", ".index", ".lookup", "_h", "_h.index", ".dbtype", "_h.dbtype"]
    for file in [filename + ext for ext in outext]:
        assert os.path.isfile(file)
    assert "Creating database" in caplog.text
    assert "Existing files: 0" in caplog.text
    assert "Expected extensions: 7" in caplog.text
    assert caplog.records[0].levelname == "INFO"
    assert caplog.records[1].levelname == "DEBUG"
    assert caplog.records[2].levelname == "DEBUG"
    assert os.path.isfile(logfile)


def test_create_mmseqdb_exist(caplog):
    """
    Check that, when trying to create mmseqdb while all output files already exist,
    it logs a warning message and quits without creating it
    """
    filename = os.path.join(GENEPATH, "test_create_mmseqsdb_exist.msdb")
    outext = ["", ".index", ".lookup", "_h", "_h.index", ".dbtype", "_h.dbtype"]
    for file in [filename + ext for ext in outext]:
        open(file, "w").close()
    prt_path = os.path.join(PATH_EXP_FILES, "exp_EXEM.All.prt")
    logfile = os.path.join(GENEPATH, "test_create_mmseqsdb_exist.log")
    mmseqs.create_mmseqs_db(filename, prt_path, logfile)

    # Check files created/existing
    for file in [filename + ext for ext in outext]:
        assert os.path.isfile(file)
    assert ("mmseqs database "
            "test/data/pangenome/generated_by_unit-tests/test_create_mmseqsdb_exist.msdb "
            "already exists. The program will use it.") in caplog.text
    assert caplog.records[0].levelname == "WARNING"


def test_create_mmseqdb_not_all_exist(caplog):
    """
    Check that, when trying to create mmseqdb while the output database exists but at least
    1 associated file is missing (here, ".dbtype") -> message saying that files already exist, except some associated so database will be recreated.
    """
    caplog.set_level(logging.DEBUG)
    filename = os.path.join(GENEPATH, "test_create_mmseqsdb_not-all-exists.msdb")
    outext_miss = ["", ".index", ".lookup", "_h.index", ".dbtype"]
    for file in [filename + ext for ext in outext_miss]:
        open(file, "w").close()
    prt_path = os.path.join(PATH_EXP_FILES, "exp_EXEM.All.prt")
    logfile = os.path.join(GENEPATH, "test_create_mmseqsdb_exist.log")

    # Run mmseqs creation
    mmseqs.create_mmseqs_db(filename, prt_path, logfile)

    # Check that all expected files have been created (not only the files created before running)
    # and remove them, as well as the logfile
    outext_exp = ["", ".index", ".lookup", "_h", "_h.index", "_h.dbtype"]
    for file in [filename + ext for ext in outext_exp]:
        assert os.path.isfile(file)

    # Get logs info
    found_text = caplog.text
    # Check log content + level
    assert ("mmseqs database test/data/pangenome/generated_by_unit-tests/"
            "test_create_mmseqsdb_not-all-exists.msdb already exists, but at least 1 "
            "associated file (.dbtype, .index etc). is missing. The program will "
            "remove existing files and recreate the database.") in found_text
    assert caplog.records[0].levelname == "WARNING"
    assert ("Removing 'test/data/pangenome/generated_by_unit-tests/"
            "test_create_mmseqsdb_not-all-exists.msdb'") in found_text
    assert ("Removing 'test/data/pangenome/generated_by_unit-tests/"
            "test_create_mmseqsdb_not-all-exists.msdb.index'.") in found_text
    assert ("Removing 'test/data/pangenome/generated_by_unit-tests/"
            "test_create_mmseqsdb_not-all-exists.msdb.lookup'") in found_text
    assert not ("Removing test/data/pangenome/generated_by_unit-tests/"
                "test_create_mmseqsdb_not-all-exists.msdb_h'.") in found_text
    assert ("Removing 'test/data/pangenome/generated_by_unit-tests/"
            "test_create_mmseqsdb_not-all-exists.msdb_h.index'.") in found_text
    assert ("Removing 'test/data/pangenome/generated_by_unit-tests/"
            "test_create_mmseqsdb_not-all-exists.msdb.dbtype'.") in found_text
    assert not ("Removing 'test/data/pangenome/generated_by_unit-tests/"
                "test_create_mmseqsdb_not-all-exists.msdb_h.dbtype'.") in found_text
    for num in range(1, 6):
        assert caplog.records[num].levelname == "DETAIL"
    assert ("Creating database") in found_text
    assert caplog.records[6].levelname == "INFO"
    assert ("Existing files: 0") in found_text
    assert caplog.records[7].levelname == "DEBUG"
    assert ("Expected extensions: 7") in found_text
    assert caplog.records[8].levelname == "DEBUG"


def test_cluster2file():
    """
    Check that given clusters are written as expected to output file
    """
    fileout = os.path.join(GENEPATH, "test_clusters2file.txt")
    clusters = {"ESCO.0216.00001.i001_00006":
                ["ESCO.0216.00002.b010_00115", "ESCO.0216.00001.i001_00006",
                 "ESCA.0216.00001.i001_00015", "ESCO.0216.00001.b010_00115",
                 "ESCO.0216.00002.i001_00300", "ESCA.0216.00001.b001_00003"],
                "ESCO.0216.00001.i001_00018":
                ["ESCO.0216.00002.b010_00130", "ESCO.0216.00001.i001_00018",
                 "ESCA.0216.00001.i001_00950", "ESCO.0216.00002.i001_00300",
                 "ESCO.0216.00001.b001_00003"],
                "ESCO.0216.00002.b010_01265":
                ["ESCO.0216.00002.b010_01265"]}
    fams = mmseqs.clusters_to_file(clusters, fileout)
    # num of fams are random. Just check that families are the same,
    # and that nums correspond to range(1, nb_families+1)
    exp_fams = [["ESCO.0216.00002.b010_01265"],
                ["ESCA.0216.00001.i001_00950", "ESCO.0216.00001.b001_00003",
                 "ESCO.0216.00001.i001_00018", "ESCO.0216.00002.b010_00130",
                 "ESCO.0216.00002.i001_00300"],
                ["ESCA.0216.00001.b001_00003", "ESCA.0216.00001.i001_00015",
                 "ESCO.0216.00001.i001_00006", "ESCO.0216.00001.b010_00115",
                 "ESCO.0216.00002.b010_00115", "ESCO.0216.00002.i001_00300"]
                ]
    # Check that families given are as expected
    for fam in list(fams.values()):
        found = False
        for exp_fam in exp_fams:
            if set(fam) == set(exp_fam):
                found = True
                break
        assert found
    # Check family numbers are as expected
    assert list(fams.keys()) == list(range(1, 4))
    # Check pangenome file
    with open(fileout, "r") as fo:
        for line in fo:
            num = int(line.split()[0])
            fam = line.split()[1:]
            assert num in list(range(1, 4))
            assert fam in exp_fams


def test_tsv2cluster():
    """
    Check that conversion from mmseq tsv file to clusters is as expected.
    """
    filein = os.path.join(PATH_TEST_FILES, "mmseq_clust-out.tsv")
    clusters = mmseqs.mmseq_tsv_to_clusters(filein)
    for res, clust in clusters.items():
        found = False
        for resx, clustx in EXP_CLUSTERS.items():
            if resx == res and set(clust) == set(clustx):
                found = True
                break
        assert found


def test_tsv2pangenome_outgiven():
    """
    From mmseq tsv file, generate output pangenome file with a given name for it
    """
    mmseqclust = os.path.join(PATH_TEST_FILES, "mmseq_clust-out")
    logmmseq = os.path.join(GENEPATH, "test_tsv2pan.log")
    start = time.strftime('%Y-%m-%d_%H-%M-%S')
    outfile1 = os.path.join(GENEPATH, "test_tsv2pan_outpangenome.txt")
    fams, outfile = mmseqs.mmseqs_tsv_to_pangenome(mmseqclust, logmmseq, start, outfile1)
    # Check that outfile nae was not modified by the function
    assert outfile1 == outfile
    # Check that the number of families is as expected
    assert len(fams) == 16
    # Check that all families found are expected families
    # + that family numbers are between 1 and nb_families (same number of families found, and
    # consistent family numbers)
    for num, fam in fams.items():
        assert num in list(range(1, 17))
        found = False
        for expfam in list(EXP_CLUSTERS.values()):
            if fam == expfam:
                found = True
                break
        assert found
    # Check that all families written in output file are as expected output file
    # (with exception that familiy numbers are not necessarily in the same order)
    exp_pan = os.path.join(PATH_EXP_FILES, "exp_pangenome-4genomes.lst")
    with open(exp_pan, "r") as ep, open(outfile, "r") as pan:
        lines_exp = []
        lines_out = []
        for line_exp, line in zip(ep, pan):
            lines_exp += line_exp.split()[1:]
            lines_out += line.split()[1:]
    assert set(lines_exp) == set(lines_out)
    # Check content of logfile
    with open(logmmseq, "r") as logf:
        first_lines = [logf.readline() for _ in range(3)]
        assert first_lines == ["\n", "------------\n", "\n"]
        start_line = logf.readline().strip()
        assert start_line == f"Start: {start}"
        end_line = logf.readline().strip()
        assert "End: " in end_line


def test_tsv2pangenome_default():
    """
    From mmseq tsv file, generate output pangenome file with default name
    """
    mmseqclust = os.path.join(PATH_TEST_FILES, "mmseq_clust-out")
    mmseqsclust_test = os.path.join(GENEPATH, "mmseq_clust-out_for_test")
    shutil.copyfile(mmseqclust, mmseqsclust_test)
    mmseqclust_tsv = os.path.join(PATH_TEST_FILES, "mmseq_clust-out.tsv")
    mmseqsclust_test_tsv = os.path.join(GENEPATH, "mmseq_clust-out_for_test.tsv")
    shutil.copyfile(mmseqclust_tsv, mmseqsclust_test_tsv)
    logmmseq = os.path.join(GENEPATH, "test_tsv2pan-def.log")
    start = time.strftime('%Y-%m-%d_%H-%M-%S')
    fams, outfile = mmseqs.mmseqs_tsv_to_pangenome(mmseqsclust_test, logmmseq, start)
    exp_out = os.path.join(GENEPATH, "PanGenome-mmseq_clust-out_for_test.tsv.lst")
    # Check name of output file (because not given)
    assert outfile == exp_out
    # Check that the number of families is as expected
    assert len(fams) == 16
    # Check that families returned are as expected
    for num, fam in fams.items():
        assert num in list(range(1, 17))
        found = False
        for expfam in list(EXP_CLUSTERS.values()):
            if fam == expfam:
                found = True
                break
        assert found
    # Check output file contains expected families
    exp_pan = os.path.join(PATH_EXP_FILES, "exp_pangenome-4genomes.lst")
    with open(exp_pan, "r") as ep, open(outfile, "r") as pan:
        lines_exp = []
        lines_out = []
        for line_exp, line in zip(ep, pan):
            lines_exp += line_exp.split()[1:]
            lines_out += line.split()[1:]
    assert set(lines_exp) == set(lines_out)
    # Check logfile content
    with open(logmmseq, "r") as logf:
        first_lines = [logf.readline() for _ in range(3)]
        assert first_lines == ["\n", "------------\n", "\n"]
        start_line = logf.readline().strip()
        assert start_line == "Start: {}".format(start)
        end_line = logf.readline().strip()
        assert "End: " in end_line


def test_mmseq2pan_givenout():
    """
    From mmseq clust output, convert to pangenome (with steps inside, already tested by the other
    functions called).+ write pangenome to ouput file
    """
    # file witch will contain pangenome
    outfile1 = os.path.join(GENEPATH, "test_mmseq2pan.lst")
    # files with output of mmseqs clusters (in mmseqs format)
    mmseqclust = os.path.join(PATH_TEST_FILES, "mmseq_clust-out")
    # mmseq db used for clustering
    mmseqdb = os.path.join(PATH_TEST_FILES, "mmseq_db")
    # Initialyze logfile with start time
    start = time.strftime('%Y-%m-%d_%H-%M-%S')
    logmmseq = os.path.join(GENEPATH, "test_mmseq2pan-out.log")

    # Run mmseqs2pan
    fams, outf = mmseqs.mmseqs_to_pangenome(mmseqdb, mmseqclust, logmmseq, start, outfile1)

    # Check output filename was not changed
    assert outfile1 == outf
    # Check that the number of families is as expected
    assert len(fams) == 16
    # assert output filename was not changed
    for num, fam in fams.items():
        # Check that the number of families return by the function is as expected
        assert num in list(range(1, 17))
        found = False
        # Check that all expected families are found in fams
        for expfam in list(EXP_CLUSTERS.values()):
            if fam == expfam:
                found = True
                break
        assert found
    exp_pan = os.path.join(PATH_EXP_FILES, "exp_pangenome-4genomes.lst")
    #Check that families written in output file are as expected
    with open(exp_pan, "r") as ep, open(outf, "r") as pan:
        lines_exp = []
        lines_out = []
        for line_exp, line in zip(ep, pan):
            lines_exp.append(tuple(line_exp.split()[1:]))
            lines_out.append(tuple(line.split()[1:]))
    assert set(lines_exp) == set(lines_out)


def test_run_clust():
    """
    Checks that, when we run mmseq clust, it creates all files needed for after to do
    the pangenome. We do not check the content of the mmseq output files, as it could
    depend on its version, and we are here testing PanACoTA.
    """
    mmseqdb = os.path.join(PATH_TEST_FILES, "mmseq_db")
    mmseqclust = os.path.join(GENEPATH, "test_mmseq_cluster-out")
    tmpdir = os.path.join(GENEPATH, "test_mmseq_tmp")
    os.makedirs(tmpdir)
    logmmseq = os.path.join(GENEPATH, "test_mmseq_cluster.log")
    min_id = 0.8
    threads = 1
    clust_mode = 1
    args = (mmseqdb, mmseqclust, tmpdir, logmmseq, min_id, threads, clust_mode)
    # Check that output of mmseq does not already exist
    assert not os.path.isfile(mmseqclust)
    # Run run_mmseqs_clust on previous arguments
    mmseqs.run_mmseqs_clust(args)
    # Check that all expected files and temporary directory are created
    # and check that no more outfile is created, in order to remove all of them !! !!
    generated_outfiles = glob.glob(mmseqclust + "*")
    assert len(generated_outfiles) == 3
    assert set(generated_outfiles) == set([mmseqclust, mmseqclust + ".index",
                                           mmseqclust + ".dbtype"])
    assert os.path.isfile(logmmseq)
    assert os.path.isdir(tmpdir)


def test_get_logmmseq():
    """
    Check that the given log filename is as expected according to given information
    """
    outdir = "toto"
    prt_bank = "bank_prt"
    infoname = "GENO115"
    log = mmseqs.get_logmmseq(outdir, prt_bank, infoname)
    assert log == "toto/mmseq_bank_prt_GENO115.log"


def test_get_info():
    """
    Check that string given by get_info is as expected according to info given in input
    """
    threads = 1
    min_id = 0.8
    clust_mode = 1
    start = "STARTTIME"
    info = mmseqs.get_info(threads, min_id, clust_mode, start)
    assert info == "0.8-mode1_STARTTIME"


def test_get_info_parallel():
    """
    Check that string given by get_info is as expected according to info given in input
    """
    threads = 12
    min_id = 0.8
    clust_mode = 1
    start = "STARTTIME"
    info = mmseqs.get_info(threads, min_id, clust_mode, start)
    assert info == "0.8-mode1-th12_STARTTIME"


def test_do_pangenome_defaultname(caplog):
    """
    Check that expected output files are created,
    and compare output pangenome to the expected one.
    """
    caplog.set_level(logging.DEBUG)
    outdir = os.path.join(GENEPATH, "test_do_pangenome_outdir")
    prt_bank = "exp_EXEM.All.prt"
    mmseqdb = os.path.join(PATH_TEST_FILES, "mmseq_db")
    min_id = 0.8
    clust_mode = 1
    threads = 1
    start = "STARTTIME"
    quiet = False
    assert not os.path.isdir(outdir)
    fams, outfile = mmseqs.do_pangenome(outdir, prt_bank, mmseqdb, min_id,
                                        clust_mode, threads, start, quiet=quiet)
    # Check creation of output directory
    assert os.path.isdir(outdir)
    # Check creation of tmp directory
    tmp_dir = os.path.join(outdir, "tmp_exp_EXEM.All.prt_0.8-mode1_STARTTIME")
    assert os.path.isdir(tmp_dir)
    # Check presence of pangenome file
    exp_out = os.path.join(outdir, "PanGenome-exp_EXEM.All.prt-clust-0.8-mode1_STARTTIME.tsv.lst")
    assert exp_out == outfile
    assert os.path.isfile(outfile)
    # Check families returned
    assert len(fams) == 16
    for num, fam in fams.items():
        assert num in list(range(1, 17))
        found = False
        for expfam in list(EXP_CLUSTERS.values()):
            if fam == expfam:
                found = True
                break
        assert found
    # Check content of output pangenome file
    exp_pan = os.path.join(PATH_EXP_FILES, "exp_pangenome-4genomes.lst")
    with open(exp_pan, "r") as ep, open(outfile, "r") as pan:
        lines_exp = []
        lines_out = []
        for line_exp, line in zip(ep, pan):
            lines_exp.append(tuple(line_exp.split()[1:]))
            lines_out.append(tuple(line.split()[1:]))
    assert set(lines_exp) == set(lines_out)
    assert "Clustering proteins..." in caplog.text
    assert caplog.records[1].levelname == "INFO"


def test_do_pangenome_given_panfile(caplog):
    """
    Check that expected output files are created,
    and compare output pangenome to the expected one.
    No possibility to check if quiet was not applied...
    """
    caplog.set_level(logging.DEBUG)
    outdir = os.path.join(GENEPATH, "test_do_pangenome_outdir_pan-name")
    prt_bank = "exp_EXEM.All.prt"
    mmseqdb = os.path.join(PATH_TEST_FILES, "mmseq_db")
    min_id = 0.8
    clust_mode = 1
    threads = 1
    start = "STARTTIME"
    quiet = False
    panfile = os.path.join(GENEPATH, "test_res_pangenome_given-name")
    # Check outdir does not exit before
    assert not os.path.isdir(outdir)
    fams, outfile = mmseqs.do_pangenome(outdir, prt_bank, mmseqdb, min_id,
                                        clust_mode, threads, start, quiet=quiet, panfile=panfile)
    # Check creation of output directory
    assert os.path.isdir(outdir)
    # Check creation of tmp directory
    tmp_dir = os.path.join(outdir, "tmp_exp_EXEM.All.prt_0.8-mode1_STARTTIME")
    assert os.path.isdir(tmp_dir)
    # Check presence of pangenome file
    assert panfile == outfile
    assert os.path.isfile(outfile)
    # Check families returned
    assert len(fams) == 16
    for num, fam in fams.items():
        assert num in list(range(1, 17))
        found = False
        for expfam in list(EXP_CLUSTERS.values()):
            if fam == expfam:
                found = True
                break
        assert found
    # Check content of output pangenome file
    exp_pan = os.path.join(PATH_EXP_FILES, "exp_pangenome-4genomes.lst")
    with open(exp_pan, "r") as ep, open(outfile, "r") as pan:
        lines_exp = []
        lines_out = []
        for line_exp, line in zip(ep, pan):
            lines_exp.append(tuple(line_exp.split()[1:]))
            lines_out.append(tuple(line.split()[1:]))
    assert set(lines_exp) == set(lines_out)
    # Check log
    assert "Clustering proteins..." in caplog.text
    assert caplog.records[1].levelname == "INFO"


def test_do_pangenome_quiet(caplog):
    """
    Check that expected output files are created,
    and compare output pangenome to the expected one.
    Check that no error appears when choosing quiet option.
    No possibility to check if quiet was applied...
    """
    caplog.set_level(logging.DEBUG)
    outdir = os.path.join(GENEPATH, "test_do_pangenome_outdir_quiet")
    prt_bank = "exp_EXEM.All.prt"
    mmseqdb = os.path.join(PATH_TEST_FILES, "mmseq_db")
    min_id = 0.8
    clust_mode = 1
    threads = 1
    start = "STARTTIME"
    quiet = True
    assert not os.path.isdir(outdir)
    fams, outfile = mmseqs.do_pangenome(outdir, prt_bank, mmseqdb, min_id,
                                        clust_mode, threads, start, quiet=quiet)
    # Check creation of output directory
    assert os.path.isdir(outdir)
    # Check creation of tmp directory
    tmp_dir = os.path.join(outdir, "tmp_exp_EXEM.All.prt_0.8-mode1_STARTTIME")
    assert os.path.isdir(tmp_dir)
    # Check presence of pangenome file
    exp_out = os.path.join(outdir, "PanGenome-exp_EXEM.All.prt-clust-0.8-mode1_STARTTIME.tsv.lst")
    assert exp_out == outfile
    assert os.path.isfile(outfile)
    # Check families returned
    for num, fam in fams.items():
        assert num in list(range(1, 17))
        found = False
        for expfam in list(EXP_CLUSTERS.values()):
            if fam == expfam:
                found = True
                break
        assert found
    # Check content of output pangenome file
    exp_pan = os.path.join(PATH_EXP_FILES, "exp_pangenome-4genomes.lst")
    with open(exp_pan, "r") as ep, open(outfile, "r") as pan:
        lines_exp = []
        lines_out = []
        for line_exp, line in zip(ep, pan):
            lines_exp.append(tuple(line_exp.split()[1:]))
            lines_out.append(tuple(line.split()[1:]))
    assert set(lines_exp) == set(lines_out)
    assert "Clustering proteins..." in caplog.text


def test_do_pangenome_exist(caplog):
    """
    Check that if the mmseq output file of clustering already exists, it does not
    run mmseq again, but just converts it to pangenome.
    """
    caplog.set_level(logging.DEBUG)
    outdir = os.path.join(GENEPATH, "test_do_pangenome_outdir_exist")
    prt_bank = "exp_EXEM.All.prt"
    mmseqdb = os.path.join(PATH_TEST_FILES, "mmseq_db")
    min_id = 0.8
    clust_mode = 1
    threads = 1
    start = "STARTTIME"
    # Create clustering results in outdir
    os.makedirs(outdir)
    orig_clust = os.path.join(PATH_TEST_FILES, "mmseq_clust-out")
    out_clust = os.path.join(outdir, "exp_EXEM.All.prt-clust-0.8-mode1_STARTTIME")
    shutil.copyfile(orig_clust, out_clust)
    shutil.copyfile(orig_clust + ".index", out_clust + ".index")
    fams, outfile = mmseqs.do_pangenome(outdir, prt_bank, mmseqdb, min_id,
                                        clust_mode, threads, start)
    assert ("mmseqs clustering test/data/pangenome/generated_by_unit-tests/"
            "test_do_pangenome_outdir_exist/exp_EXEM.All.prt-clust-0.8-mode1_STARTTIME "
            "already exists. The program will now convert it to a "
            "pangenome file.") in caplog.text
    # assert
    # Check creation of empty tmp directory
    tmp_dir = os.path.join(outdir, "tmp_exp_EXEM.All.prt_0.8-mode1_STARTTIME")
    assert os.path.isdir(tmp_dir)
    assert glob.glob(os.path.join(tmp_dir, "*")) == []
    # Check presence of pangenome file
    exp_out = os.path.join(outdir, "PanGenome-exp_EXEM.All.prt-clust-0.8-mode1_STARTTIME.tsv.lst")
    assert exp_out == outfile
    assert os.path.isfile(outfile)
    # Check families returned
    for num, fam in fams.items():
        assert num in list(range(1, 17))
        found = False
        for expfam in list(EXP_CLUSTERS.values()):
            if fam == expfam:
                found = True
                break
        assert found
    # Check content of output pangenome file
    exp_pan = os.path.join(PATH_EXP_FILES, "exp_pangenome-4genomes.lst")
    with open(exp_pan, "r") as ep, open(outfile, "r") as pan:
        lines_exp = []
        lines_out = []
        for line_exp, line in zip(ep, pan):
            lines_exp.append(tuple(line_exp.split()[1:]))
            lines_out.append(tuple(line.split()[1:]))
    assert set(lines_exp) == set(lines_out)
    assert "Clustering proteins..." not in caplog.text
    assert caplog.records[1].levelname == "WARNING"


def test_run_all_pangenome(caplog):
    """
    Check that, given a prt bank, it creates mmseq db, mmseq clustering, and
    outputs the expected pangenome file.
    """
    caplog.set_level(logging.DEBUG)
    min_id = 0.8
    clust_mode = 1
    outdir = os.path.join(GENEPATH, "test_run_allpangenome")
    os.makedirs(outdir)
    prt_path = os.path.join(PATH_EXP_FILES, "exp_EXEM.All.prt")
    threads = 1
    panfile = None
    quiet = False
    fams, outfile = mmseqs.run_all_pangenome(min_id, clust_mode, outdir, prt_path,
                                             threads, panfile=panfile, quiet=quiet)

    # check that tmp dir was created and not empty
    tmp_dir = os.path.join(outdir, "tmp_exp_EXEM.All.prt_0.8-mode1_*")
    assert glob.glob(os.path.join(tmp_dir, "*")) != []

    # check that pangenome file is present with expected name
    exp_out = os.path.join(outdir, "PanGenome-exp_EXEM.All.prt-clust-0.8-mode1_*")
    found_out = glob.glob(exp_out)
    assert len(found_out) == 1
    found_out = found_out[0]
    assert outfile == found_out
    assert os.path.isfile(outfile)

    # Check content of output pangenome file
    exp_pan = os.path.join(PATH_EXP_FILES, "exp_pangenome-4genomes.lst")
    # Compare lines, ignoring the 1st field (family number), and order.
    # -> families can be in any order/associated to any family number
    with open(exp_pan, "r") as ep, open(outfile, "r") as pan:
        lines_exp = []
        lines_out = []
        # get all lines in exp and out files, without family number
        for line_exp, line in zip(ep, pan):
            # line_exp.split() -> list
            # But set([list, list]) is not possible (list is not hashable).
            # -> use tuples instead of lists
            lines_exp.append(tuple(line_exp.split()[1:]))
            lines_out.append(tuple(line.split()[1:]))
    # Check that out lines are exactly the same (except order)
    assert set(lines_exp) == set(lines_out)

    # Check families returned in fams dict.
    for num, fam in fams.items():
        assert num in list(range(1, 17))
        found = False
        for expfam in FAMILIES4G:
            if fam == expfam:
                found = True
                break
        assert found

    # Check logs
    assert ("Will run MMseqs2 with:\n\t- minimum sequence identity = 80.0%\n"
            "\t- cluster mode 1") in caplog.text
    assert "Creating database" in caplog.text
    assert "Existing files: 0" in caplog.text
    assert "Expected extensions: 7" in caplog.text
    assert "Clustering proteins..." in caplog.text
    assert caplog.records[0].levelname == "INFO"
    assert caplog.records[1].levelname == "INFO"
    assert caplog.records[2].levelname == "DEBUG"
    assert caplog.records[3].levelname == "DEBUG"


def test_run_all_pangenome_givenfile_parallel(caplog):
    """
    Check that, given a prt bank, it creates mmseq db, mmseq clustering, and
    outputs the expected pangenome file.
    """
    caplog.set_level(logging.DEBUG)
    min_id = 0.8
    clust_mode = 1
    outdir = os.path.join(GENEPATH, "test_run_allpangenome_paralell")
    os.makedirs(outdir)
    prt_path = os.path.join(PATH_EXP_FILES, "exp_EXEM.All.prt")
    threads = 2
    panfile = "pangenome_test_run-all-pan.lst"
    quiet = True
    fams, outfile = mmseqs.run_all_pangenome(min_id, clust_mode, outdir, prt_path,
                                             threads, panfile=panfile, quiet=quiet)

    # check that tmp dir was created and not empty
    tmp_dir = os.path.join(outdir, "tmp_exp_EXEM.All.prt_0.8-mode1-th2*")
    assert glob.glob(os.path.join(tmp_dir, "*")) != []

    # check that pangenome file is present
    assert outfile == os.path.join(outdir, panfile)
    assert os.path.isfile(outfile)

    # Check content of output pangenome file
    exp_pan = os.path.join(PATH_EXP_FILES, "exp_pangenome-4genomes.lst")
    # Compare lines, ignoring the 1st field (family number), and order.
    # -> families can be in any order/associated to any family number
    with open(exp_pan, "r") as ep, open(outfile, "r") as pan:
        lines_exp = []
        lines_out = []
        for line_exp, line in zip(ep, pan):
            lines_exp.append(tuple(line_exp.split()[1:]))
            lines_out.append(tuple(line.split()[1:]))
    # Check that out lines are exactly the same (except order)
    assert set(lines_exp) == set(lines_out)

    # Check families returned
    for num, fam in fams.items():
        assert num in list(range(1, 17))
        found = False
        for expfam in FAMILIES4G:
            if fam == expfam:
                found = True
                break
        assert found

    # Check logs
    assert ("Will run MMseqs2 with:\n\t- minimum sequence identity = 80.0%\n"
            "\t- cluster mode 1\n\t- 2 threads") in caplog.text
    assert "Creating database" in caplog.text
    assert "Existing files: 0" in caplog.text
    assert "Expected extensions: 7" in caplog.text
    assert "Clustering proteins..." in caplog.text
    assert caplog.records[0].levelname == "INFO"
    assert caplog.records[1].levelname == "INFO"
    assert caplog.records[2].levelname == "DEBUG"
    assert caplog.records[3].levelname == "DEBUG"



#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
Functions to build a bank of all proteins to include in the pangenome

@author gem
April 2017
"""
from PanACoTA import utils
import logging
import os

logger = logging.getLogger('pangenome.bank')


def build_prt_bank(lstinfo, dbpath, name, spedir, quiet):
    """
    Build a file containing all proteins of all genomes contained in lstinfo.

    Parameters
    ----------
    lstinfo : str
        1 line per genome, only 1st column considered here, as the genome name
        without extension
    dbpath : str
        Proteins folder, containing all proteins for each genome. Each genome has
        its own protein file, called `<genome_name>.prt`.
    name : str
        dataset name, used to name the output databank: <outdir>/<name>.All.prt
    spedir : str or None
        By default, output file is saved in dbpath directory. If it must be saved somewhere
        else, it is specified here.
    quiet : bool
        True if nothing must be written in stdout/stderr, False otherwise

    Returns
    -------
    str
        name (with path) of the protein databank generated
    """
    if not spedir:
        outdir = dbpath
    else:
        os.makedirs(spedir, exist_ok=True)
        outdir = spedir
    outfile = os.path.join(outdir, name + ".All.prt")
    if os.path.isfile(outfile):
        logger.warning((f"Protein bank {outfile} already exists. "
                        "It will be used by mmseqs."))
        return outfile
    logger.info(f"Building bank with all proteins to {outfile}")
    genomes = []
    with open(lstinfo) as lstf:
        for line in lstf:
            # skip header
            if "_name" in line:
                continue
            genome = line.strip().split()[0].strip()
            genomes.append(genome)
    all_names = [os.path.join(dbpath, gen + ".prt") for gen in genomes]
    if quiet:
        utils.cat(all_names, outfile)
    else:
        utils.cat(all_names, outfile, title="Building bank")
    return outfile
